Package: SAIGEgds
Type: Package
Title: Scalable Implementation of Generalized mixed models using GDS
        files in Phenome-Wide Association Studies
Version: 2.12.0
Date: 2026-04-23
Depends: R (>= 4.0.0), gdsfmt (>= 1.28.0), SeqArray (>= 1.50.2), Rcpp
LinkingTo: Rcpp, RcppArmadillo, RcppParallel (>= 5.0.0)
Imports: methods, stats, utils, Matrix, RcppParallel, SKAT,
        CompQuadForm, survey
Suggests: parallel, markdown, rmarkdown, crayon, SNPRelate, RUnit,
        knitr, ggmanh, BiocGenerics
Authors@R: c(person("Xiuwen", "Zheng", role=c("aut", "cre"),
        email="xiuwen.zheng@abbvie.com", comment=c(ORCID="0000-0002-1390-0708")),
        person("Wei", "Zhou", role="ctb",
        comment="the original author of the SAIGE R package"),
        person("J. Wade", "Davis", role="ctb"))
Description: Scalable implementation of generalized mixed models with highly
        optimized C++ implementation and integration with Genomic Data Structure
        (GDS) files. It is designed for single variant tests and set-based
        aggregate tests in large-scale Phenome-wide Association Studies (PheWAS)
        with millions of variants and samples, controlling for sample structure
        and case-control imbalance. The implementation is based on the SAIGE R
        package (v0.45, Zhou et al. 2018 and Zhou et al. 2020), and it is
        extended to include the state-of-the-art ACAT-O set-based tests.
        Benchmarks show that SAIGEgds is significantly faster than the SAIGE R
        package. Optional OpenCL-based GPU acceleration is supported for the
        GRM cross-product computation in null model fitting and for GRM
        construction.
License: GPL-3
SystemRequirements: GNU make
VignetteBuilder: knitr
ByteCompile: TRUE
URL: https://github.com/AbbVie-ComputationalGenomics/SAIGEgds
biocViews: Software, Genetics, StatisticalMethod, GenomeWideAssociation
git_url: https://git.bioconductor.org/packages/SAIGEgds
git_branch: RELEASE_3_23
git_last_commit: d1163ac
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-28 23:22:39 UTC; biocbuild
Author: Xiuwen Zheng [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1390-0708>),
  Wei Zhou [ctb] (the original author of the SAIGE R package),
  J. Wade Davis [ctb]
Maintainer: Xiuwen Zheng <xiuwen.zheng@abbvie.com>
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-29 04:58:52 UTC; unix

# Export the following names
useDynLib(SAIGEgds,
    saige_simd_version, saige_simd_sp_grm, saige_grm_ds_calc,
    saige_set_numthread,
    saige_init_sparse, saige_get_sparse_info,
    saige_init_fit_grm, saige_store_2b_geno, saige_store_sp_geno,
    saige_store_dense_grm, saige_store_sparse_grm, saige_get_grm_diag,
    saige_score_test_init,
    saige_fit_AI_PCG, saige_calc_var_ratio, saige_set_geno2b_raw,
    saige_score_test_pval,
    saige_burden_test_pval, saige_acatv_test_pval,
    saige_skat_test_init, saige_skat_test_reset, saige_skat_test_done,
    saige_skat_test_pval,
    saige_acato_test_pval,
    saige_acat_p,
    saige_grm_sp_calc, saige_grm_sp_calc_ijx, saige_grm_sp_reraw,
    saige_set_option,
    saige_get_G0_flipped_impute,
    saige_gpu_init, saige_gpu_cleanup
)

# Export all names
exportPattern("^seq")
export(pACAT, pACAT2, glmmHeritability)

# Import all packages listed as Imports or Depends
import(methods)
import(Rcpp)
import(Matrix)
importFrom(stats, as.formula, binomial, gaussian, glm, lm, model.frame,
    model.matrix, model.response, na.omit, sd, var, qnorm, dbeta)
importFrom(utils, object.size, packageVersion, str, write.csv, write.table)
importFrom(gdsfmt, createfn.gds, openfn.gds, closefn.gds, index.gdsn, ls.gdsn,
    add.gdsn, read.gdsn, readex.gdsn, readmode.gdsn, put.attr.gdsn,
    get.attr.gdsn, exist.gdsn, append.gdsn, moveto.gdsn, cleanup.gds)
importFrom(SeqArray, seqOpen, seqClose, seqGetData, seqApply, seqSummary,
    seqParallel, seqResetFilter, seqSetFilter, seqGetFilter, seqSetFilterCond,
    seqAlleleFreq, seqBlockApply, seqUnitApply, seqGet2bGeno, seqMerge,
    seqGetAF_AC_Missing, seqExport, seqFilterPush, seqFilterPop, seqParApply)
importFrom(RcppParallel, setThreadOptions, RcppParallelLibs)
importFrom(CompQuadForm, davies)

# Registering S3 methods
S3method(print, ClassSAIGE_NullModel)

SAIGEgds: Scalable Implementation of Generalized mixed models in PheWAS using GDS files
====

![GPLv3](http://www.gnu.org/graphics/gplv3-88x31.png)
[GNU General Public License, GPLv3](http://www.gnu.org/copyleft/gpl.html)


## Features

Scalable implementation of generalized mixed mode with the support of Genomic Data Structure ([GDS](https://github.com/zhengxwen/SeqArray)) files and highly optimized C++ implementation. It is designed for single variant tests in large-scale phenome-wide association studies (PheWAS) with millions of variants and hundreds of thousands of samples (e.g., [UK Biobank genotype data](https://www.ukbiobank.ac.uk/scientists-3/genetic-data)), controlling for case-control imbalance and sample structure in single variant association studies.

The implementation of SAIGEgds is based on the original [SAIGE](https://github.com/weizhouUMICH/SAIGE) R package (v0.29.4.4) [Zhou et al. 2018]. It is implemented with optimized C++ codes taking advantage of sparse structure of genotypes. All of the calculation with single-precision floating-point numbers in [SAIGE](https://github.com/weizhouUMICH/SAIGE) are replaced by the double-precision calculation in SAIGEgds. SAIGEgds also implements some of the [SPAtest](https://cran.r-project.org/web/packages/SPAtest/index.html) functions in C to speed up the calculation of Saddlepoint approximation.

Benchmarks using the UK Biobank White British genotype data (N=430K) with coronary heart disease and simulated cases, show that SAIGEgds is 5 to 6 times faster than the SAIGE R package in the steps of fitting null models and p-value calculations. When used in conjunction with high-performance computing (HPC) clusters and/or cloud resources, SAIGEgds provides an efficient analysis pipeline for biobank-scale PheWAS.


## Bioconductor:

Release Version: v2.12.0 ([http://www.bioconductor.org/packages/SAIGEgds](http://www.bioconductor.org/packages/SAIGEgds))

* [Help Documents](https://rdrr.io/bioc/SAIGEgds/man)
* [Tutorial](http://www.bioconductor.org/packages/release/bioc/vignettes/SAIGEgds/inst/doc/SAIGEgds.html)
* [News](http://www.bioconductor.org/packages/release/bioc/news/SAIGEgds/NEWS)


## Package Maintainer

[Xiuwen Zheng](xiuwen.zheng@abbvie.com)


## Installation

* Requires R (≥ v3.5.0), [gdsfmt](http://www.bioconductor.org/packages/gdsfmt) (≥ v1.20.0), [SeqArray](http://www.bioconductor.org/packages/SeqArray) (≥ v1.32.0)

* Recommend [GNU GCC (≥ v6.0)](https://gcc.gnu.org), requiring C++11

* Bioconductor repository
```R
if (!requireNamespace("BiocManager", quietly=TRUE))
    install.packages("BiocManager")
BiocManager::install("SAIGEgds")
```
The `BiocManager::install()` approach may require that you build from source, i.e. `make` and compilers must be installed on your system -- see the [R FAQ](http://cran.r-project.org/faqs.html) for your operating system; you may also need to install dependencies manually.

* Development version from Github (for developers/testers only)
```R
library("devtools")
install_github("AbbVie-ComputationalGenomics/SAIGEgds")
```


## Package vignette

If the package is installed from Bioconductor repository or package rebuilding, users can start R and enter to view documentation:
```R
browseVignettes("SAIGEgds")
```


## Examples

```R
library(SeqArray)
library(SAIGEgds)

# open the GDS file for genetic relationship matrix (GRM)
grm_fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
(grm_gds <- seqOpen(grm_fn))

# load phenotype
phenofn <- system.file("extdata", "pheno.txt.gz", package="SAIGEgds")
pheno <- read.table(phenofn, header=TRUE, as.is=TRUE)
head(pheno)
##   sample.id y     yy      x1 x2
## 1        s1 0 4.5542  1.5118  1
## 2        s2 0 3.7941  0.3898  1
## 3        s3 0 5.0411 -0.6212  1
## ...

# fit the null model
glmm <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, grm_gds, trait.type="binary",
    sample.col="sample.id", num.thread=2)
## SAIGE association analysis:
## Filtering variants:
## [==================================================] 100%, completed, 0s
## Fit the null model: y ~ x1 + x2 + var(GRM)
##     # of samples: 1,000
##     # of variants: 9,976
##     using 2 threads
## ...

# close the file
seqClose(grm_gds)



################################

# open the GDS file for association testing
geno_fn <- system.file("extdata", "assoc_100snp.gds", package="SAIGEgds")
(geno_gds <- seqOpen(geno_fn))
## File: assoc_100snp.gds (10.5K)
## +    [  ] *
## |--+ description   [  ] *
## |--+ sample.id   { Str8 1000 LZMA_ra(12.6%), 625B }
## |--+ variant.id   { Int32 100 LZMA_ra(48.5%), 201B } *
## ...


# p-value calculation
assoc <- seqAssocGLMM_SPA(geno_gds, glmm, mac=10, parallel=2)
## SAIGE association analysis:
##     # of samples: 1,000
##     # of variants: 100
##     MAF threshold: NaN
##     MAC threshold: 10
##     missing threshold for variants: 0.1
##     p-value threshold for SPA adjustment: 0.05
##     variance ratio for approximation: 0.9391186
##     # of processes: 2
## [==================================================] 100%, completed, 0s
## # of variants after filtering by MAF, MAC and missing thresholds: 38
## Done.

head(assoc)
##   id chr pos rs.id ref alt AF.alt mac  num       beta        SE      pval pval.noadj converged
## 1  4   1   4   rs4   A   C 0.0100  20 1000  -0.074992  0.791685  0.924533   0.924533      TRUE
## 2 12   1  12  rs12   A   C 0.0150  30 1000  -0.091001  0.657140  0.889861   0.889861      TRUE
## 3 14   1  14  rs14   A   C 0.0375  75 1000  -0.075455  0.434152  0.862023   0.862023      TRUE
## ...


# close the file
seqClose(geno_gds)
```


## Citations

Zheng X, Davis J.Wade. SAIGEgds -- an efficient statistical tool for large-scale PheWAS with mixed models. *Bioinformatics* (2021). Mar;37(5):728-730. [DOI: 10.1093/bioinformatics/btaa731](http://dx.doi.org/10.1093/bioinformatics/btaa731).

Zhou W, Zhao Z, Nielsen JB, Fritsche LG, LeFaive J, Taliun SAG, Bi W, Gabrielsen ME, Daly MJ, Neale BM, Hveem K, Abecasis GR, Willer CJ, Lee S. Scalable generalized linear mixed model for region-based association tests in large biobanks and cohorts. *Nat Genet* (2020). Jun;52(6):634-639. [DOI:10.1038/s41588-020-0621-6](http://dx.doi.org/10.1038/s41588-020-0621-6)

Zhou W, Nielsen JB, Fritsche LG, Dey R, Gabrielsen ME, Wolford BN, LeFaive J, VandeHaar P, Gagliano SA, Gifford A, Bastarache LA, Wei WQ, Denny JC, Lin M, Hveem K, Kang HM, Abecasis GR, Willer CJ, Lee S. Efficiently controlling for case-control imbalance and sample relatedness in large-scale genetic association studies. *Nat Genet* (2018). Sep;50(9):1335-1341. [DOI: 10.1038/s41588-018-0184-y](https://www.nature.com/articles/s41588-018-0184-y).

Zheng X, Gogarten S, Lawrence M, Stilp A, Conomos M, Weir BS, Laurie C, Levine D. SeqArray -- A storage-efficient high-performance data format for WGS variant calls. *Bioinformatics* (2017). [DOI: 10.1093/bioinformatics/btx145](http://dx.doi.org/10.1093/bioinformatics/btx145).


## See Also

[SeqArray](https://www.bioconductor.org/packages/SeqArray): Data management of large-scale whole-genome sequence variant calls

[gds2bgen](https://github.com/zhengxwen/gds2bgen): Format conversion from bgen to gds

#######################################################################
#
# Package name: SAIGEgds
#
# Description:
#     Scalable and accurate implementation of generalized mixed models
# using GDS files
#
# Copyright (C) 2019-2026    Xiuwen Zheng / AbbVie-ComputationalGenomics
# License: GPL-3
#


# Package-wide variable
.PkgEnv <- new.env()


#######################################################################
# Internal functions
#

.set_option <- function(always_fastSPA=TRUE, use_avx=c("avx512f","avx2","no"),
    verbose=TRUE)
{
    use_avx <- match.arg(use_avx)
    use_avx <- match(use_avx, c("avx512f","avx2","no"))
    .Call(saige_set_option, always_fastSPA, use_avx, verbose)
    invisible()
}

.cfunction <- function(name)
{
    fn <- function(x) NULL
    f <- quote(.Call(SEQ_ExternalName1, x))
    f[[1L]] <- .Call
    f[[2L]] <- getNativeSymbolInfo(name, "SAIGEgds")$address
    body(fn) <- f
    fn
}

.cfunction2 <- function(name)
{
    fn <- function(x, y) NULL
    f <- quote(.Call(SEQ_ExternalName2, x, y))
    f[[1L]] <- .Call
    f[[2L]] <- getNativeSymbolInfo(name, "SAIGEgds")$address
    body(fn) <- f
    fn
}

.cat <- function(...) cat(..., "\n", sep="")

.load_pkg <- quote({
    library(Rcpp, quietly=TRUE)
    library(SAIGEgds, quietly=TRUE)
})

.pretty <- function(x)
{
    prettyNum(x, big.mark=",", scientific=FALSE)
}

.pretty_size <- function(x)
{
    if (x >= 1024^4)
        sprintf("%.1fT", x/1024^4)
    else if (x >= 1024^3)
        sprintf("%.1fG", x/1024^3)
    else if (x >= 1024^2)
        sprintf("%.1fM", x/1024^2)
    else if (x >= 1024)
        sprintf("%.1fK", x/1024)
    else
        sprintf("%g bytes", x)
}

.pretty_gt_eq <- function(x)
{
    if (is.na(x)) "no" else sprintf(">= %.15g", x)
}

.pretty_lt_eq <- function(x)
{
    if (is.na(x)) "no" else sprintf("<= %.15g", x)
}

SIMD <- function()
{
    .Call(saige_simd_version)
}

.rank_norm <- function(y, m=0, s=1)
{
    if (anyNA(y))
        stop("NA is not allowed in the inverse normal transform.")
    qnorm((rank(y) - 0.5)/length(y), mean=m, sd=s)
}

.crayon_inverse <- function(s)
{
    if (getOption("gds.crayon", TRUE) &&
            requireNamespace("crayon", quietly=TRUE))
        s <- crayon::inverse(s)
    s
}

.crayon_underline <- function(s)
{
    if (getOption("gds.crayon", TRUE) &&
            requireNamespace("crayon", quietly=TRUE))
        s <- crayon::underline(s)
    s
}

# Write to GDS file
.write_gds <- function(out.gds, out.nm, in.gds, in.nm, cm)
{
    i <- index.gdsn(in.gds, in.nm)
    n <- add.gdsn(out.gds, out.nm, storage=i, compress=cm, replace=TRUE)
    seqApply(in.gds, in.nm, `c`, as.is=n)
    readmode.gdsn(n)
    n
}

# date string
.tm <- function() strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")

# Check null model
.check_modobj <- function(modobj, verbose)
{
    if (is.character(modobj))
    {
        stopifnot(length(modobj)==1L)
        if (verbose)
            .cat("    load the null model from ", sQuote(modobj))
        if (grepl("\\.(rda|RData)$", modobj, ignore.case=TRUE))
        {
            modobj <- get(load(modobj))
        } else if (grepl("\\.rds$", modobj, ignore.case=TRUE))
        {
            modobj <- readRDS(modobj)
        } else
            stop("It should be an RData, rda or rds file.")
    }
    stopifnot(inherits(modobj, "ClassSAIGE_NullModel"))
    modobj
}

# Show the distribution
.show_outcome <- function(trait.type, y, phenovar=NULL)
{
    if (trait.type == "binary")
    {
        # binary outcome
        if (!is.null(phenovar))
            .cat("Binary outcome: ", phenovar)
        v <- table(y)
        n <- length(v) 
        v <- data.frame(v, as.numeric(prop.table(v)))
        v[, 1L] <- paste0("      ", v[, 1L])
        colnames(v) <- c(phenovar, "Number", "Proportion")
        print(v, row.names=FALSE)
        if (n != 2L)
            stop("The outcome variable has more than 2 categories!")
    } else if (trait.type == "quantitative")
    {
        # quantitative outcome
        if (!is.null(phenovar))
            .cat("Quantitative outcome: ", phenovar)
        v <- data.frame(mean=mean(y), sd=sd(y), min=min(y), max=max(y))
        rownames(v) <- "   "
        print(v)
    }
}


# S3 print method
print.ClassSAIGE_NullModel <- function(x, ...) str(x)


#######################################################################
# p-value from Cauchy-based ACAT combination method
#

pACAT <- function(p, w=NULL)
{
    .Call(saige_acat_p, p, w)
}

pACAT2 <- function(p, maf, wbeta=c(1,25))
{
    stopifnot(is.numeric(wbeta), length(wbeta)==2L)
    stopifnot(length(p) == length(maf))
    w <- dbeta(maf, wbeta[1L], wbeta[2L])
    .Call(saige_acat_p, p, w * w * maf * (1-maf))
}



#######################################################################
# Load the association p-values in a GDS file
#

seqSAIGE_LoadPval <- function(fn, varnm=NULL, index=NULL, verbose=TRUE)
{
    # check
    stopifnot(is.character(fn), length(fn)>0L, all(!is.na(fn)))
    stopifnot(is.null(varnm) || is.character(varnm))
    stopifnot(is.null(index) || is.numeric(index) || is.logical(index))
    stopifnot(is.logical(verbose), length(verbose)==1L)

    fn <- unique(fn)
    if (length(fn) == 1L)
    {
        if (verbose)
            .cat("Loading ", sQuote(fn), " ...")
        if (grepl("\\.gds$", fn, ignore.case=TRUE))
        {
            f <- openfn.gds(fn)
            on.exit(closefn.gds(f))
            fm <- get.attr.gdsn(f$root)$FileFormat[1L]
            if (length(fm)==0L) fm <- ""
            if (fm %in% c("SAIGE_OUTPUT", "SAIGE_OUTPUT_SET"))
            {
                if (is.null(varnm))
                    varnm <- ls.gdsn(f$root)
                varnm <- setdiff(varnm, "sample.id")
                rv <- list()
                for (nm in varnm)
                    rv[[nm]] <- readex.gdsn(index.gdsn(f, nm), index)
                rv <- as.data.frame(rv, stringsAsFactors=FALSE)
            } else {
                stop("FileFormat should be ",
                    "'SAIGE_OUTPUT' or 'SAIGE_OUTPUT_BURDEN'.")
            }
        } else if (grepl("\\.(rda|RData)$", fn, ignore.case=TRUE))
        {
            rv <- get(load(fn))
            if (!is.null(varnm)) rv <- rv[, varnm]
            if (!is.null(index)) rv <- rv[index, ]
        } else if (grepl("\\.rds$", fn, ignore.case=TRUE))
        {
            rv <- readRDS(fn)
            if (!is.null(varnm)) rv <- rv[, varnm]
            if (!is.null(index)) rv <- rv[index, ]
        } else
            stop("Unknown format, should be RData, RDS or gds.")
    } else {
        if (!is.null(index))
            stop("'index' should be NULL for multiple input files.")
        rv <- lapply(fn, function(nm)
                seqSAIGE_LoadPval(nm, varnm, verbose=verbose))
        if (verbose) cat("Merging ...\n")
        rv <- do.call(rbind, rv)
        if (verbose)
            cat(sprintf("Done: %d rows merged\n", NROW(rv)))
    }
    rv
}



#######################################################################
# Heritability estimation
#

glmmHeritability <- function(modobj, adjust=TRUE)
{
    # check
    stopifnot(is.logical(adjust), length(adjust)==1L)
    modobj <- .check_modobj(modobj, FALSE)

    if (modobj$trait.type == "binary")
    {
        tau <- modobj$tau[2L]
        r <- 1
        if (isTRUE(adjust))
        {
            y <- unname(modobj$obj.noK$y)
            p <- sum(y==1) / length(y)
            # coefficients are based on Supplementary Table 7 (Zhou et al. 2018)
            r <- 2.970 + 0.372*log10(p)
        }
        h <- tau / (pi*pi/3 + tau) * r
    } else if (modobj$trait.type == "quantitative")
    {
        h <- modobj$tau[2L] / sum(modobj$tau)
    } else
        stop("Invalid 'modobj$trait.type'.")

    unname(h)
}

#############################################################
#
# DESCRIPTION: test the SAIGE calculation
#

library(RUnit)
library(SAIGEgds)


# create an R object file for package checking
.create_test_file <- function()
{
	# 'Rounding' was the default in versions prior to R_3.6.0
	# it is used for reproduction of the results created by R (v3.5.2)
	tryCatch(suppressWarnings(RNGkind("Mersenne-Twister", "Inversion", "Rounding")),
		error=function(e) FALSE)

	# open a GDS file
	fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
	gdsfile <- seqOpen(fn)
	on.exit(seqClose(gdsfile))

	# load phenotype
	phenofn <- system.file("extdata", "pheno.txt.gz", package="SAIGEgds")
	pheno <- read.table(phenofn, header=TRUE, as.is=TRUE)

	# fit the null model for binary outcomes
	glmm <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, gdsfile)
	# save model
	saveRDS(glmm, file="saige_model.rds", compress="xz")
	# p-value calculation
	seqAssocGLMM_SPA(gdsfile, glmm, mac=4, res.savefn="saige_pval.rds")

	# fit the null model for quantitative outcomes
	glmm <- seqFitNullGLMM_SPA(yy ~ x1 + x2, pheno, gdsfile,
		trait.type="quantitative")
	# save model
	saveRDS(glmm, file="saige_model_quant.rds", compress="xz")
	# p-value calculation
	seqAssocGLMM_SPA(gdsfile, glmm, mac=4, res.savefn="saige_pval_quant.rds")
}


test.saige_fit_null_model <- function()
{
	# 'Rounding' was the default in versions prior to R_3.6.0
	# it is used for reproduction of the results created by R (v3.5.2)
	tryCatch(suppressWarnings(RNGkind("Mersenne-Twister", "Inversion", "Rounding")),
		error=function(e) FALSE)

	# load the previous models
	mod1 <- readRDS(system.file("unitTests", "saige_model.rds",
		package="SAIGEgds"))
	mod2 <- readRDS(system.file("unitTests", "saige_model_quant.rds",
		package="SAIGEgds"))

	# open a GDS file
	fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
	gdsfile <- seqOpen(fn)
	on.exit(seqClose(gdsfile))

	# load phenotype
	phenofn <- system.file("extdata", "pheno.txt.gz", package="SAIGEgds")
	pheno <- read.table(phenofn, header=TRUE, as.is=TRUE)

	# fit the null model, check binary outcomes
	glmm <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, gdsfile)
	checkEquals(mod1, glmm, "check the SAIGE parameters (binary outcomes)",
		tolerance=1e-4)

	# fit the null model, check quantitative outcomes
	glmm <- seqFitNullGLMM_SPA(yy ~ x1 + x2, pheno, gdsfile,
		trait.type="quantitative")
	checkEquals(mod2, glmm,
		"check the SAIGE parameters (quantitative outcomes)", tolerance=1e-4)
}


test.saige_pval <- function()
{
	# load the previous models and results
	mod1 <- readRDS(system.file("unitTests", "saige_model.rds",
		package="SAIGEgds"))
	mod2 <- readRDS(system.file("unitTests", "saige_model_quant.rds",
		package="SAIGEgds"))
	pval1 <- readRDS(system.file("unitTests", "saige_pval.rds",
		package="SAIGEgds"))
	pval2 <- readRDS(system.file("unitTests", "saige_pval_quant.rds",
		package="SAIGEgds"))

	# open a GDS file
	fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
	gdsfile <- seqOpen(fn)
	on.exit(seqClose(gdsfile))

	# p-value calculation, check binary outcomes
	assoc <- seqAssocGLMM_SPA(gdsfile, mod1, mac=4)
	checkEquals(pval1, assoc,
		"check the SAIGE p-value output (binary outcomes)", tolerance=1e-7)

	# p-value calculation, check quantitative outcomes
	assoc <- seqAssocGLMM_SPA(gdsfile, mod2, mac=4)
	checkEquals(pval2, assoc,
		"check the SAIGE p-value output (quantitative outcomes)",
		tolerance=1e-7)
}


test.saige_acta_o <- function()
{
	# load the prefit model
	mod <- readRDS(system.file("unitTests", "saige_model.rds",
		package="SAIGEgds"))

	# open a GDS file
	fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
	gdsfile <- seqOpen(fn)
	on.exit(seqClose(gdsfile))

	# get a list of variant units for aggregate tests
	ut <- seqUnitSlidingWindows(gdsfile, win.size=200, win.shift=100)

	# run burden, ACAT-V & ACAT-O
	o <- seqAssocGLMM_spaACAT_O(gdsfile, mod, ut)
	v <- seqAssocGLMM_spaACAT_V(gdsfile, mod, ut)
	b <- seqAssocGLMM_spaBurden(gdsfile, mod, ut)

	# check p-value
	checkEquals(o$pval.b1_1, b$pval.b1_1, "ACAT-O vs Burden, beta(1,1)")
	checkEquals(o$pval.b1_25, b$pval.b1_25, "ACAT-O vs Burden, beta(1,25)")
	checkEquals(o$pval.v1_1, v$pval.v1_1, "ACAT-O vs ACAT-V, beta(1,1)")
	checkEquals(o$pval.v1_25, v$pval.v1_25, "ACAT-O vs ACAT-V, beta(1,25)")
}


test.pACAT <- function()
{
    # R implementation
    ps <- 10^-seq(1, 15, 0.1)
    A1 <- matrix(0, nrow=length(ps), ncol=length(ps))
    for (i in seq_along(ps))
    {
        for (j in seq_along(ps))
        {
            T <- mean(c(tanpi(0.5 - ps[i]), tanpi(0.5 - ps[j])))
            A1[i, j] <- 0.5 - atan(T)/pi
        }
    }

    # C implementation
    A2 <- matrix(0, nrow=length(ps), ncol=length(ps))
    for (i in seq_along(ps))
        for (j in seq_along(ps))
            A2[i, j] <- pACAT(c(ps[i], ps[j]))

    # check
	checkEquals(A1, A2, "R / C implementation of ACAT")
}


test.saige_gpu_fallback <- function()
{
	# This test verifies that use.gpu=TRUE works without error
	# even when no GPU is available (graceful fallback to CPU).

	# open a GDS file
	fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
	gdsfile <- seqOpen(fn)
	on.exit(seqClose(gdsfile))

	# load phenotype
	phenofn <- system.file("extdata", "pheno.txt.gz", package="SAIGEgds")
	pheno <- read.table(phenofn, header=TRUE, as.is=TRUE)

	# fit the null model with use.gpu=TRUE (should fall back to CPU gracefully)
	glmm_cpu <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, gdsfile,
		use.gpu=FALSE, verbose=FALSE)
	glmm_gpu <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, gdsfile,
		use.gpu=TRUE, verbose=FALSE)

	# results should match (if GPU was used, within tolerance; if CPU fallback, exact)
	checkEquals(glmm_cpu$tau, glmm_gpu$tau,
		"GPU fallback: tau should match CPU", tolerance=1e-6)
	checkEquals(glmm_cpu$coefficients, glmm_gpu$coefficients,
		"GPU fallback: coefficients should match CPU", tolerance=1e-6)
}

#######################################################################
#
# Package name: SAIGEgds
#
# Description:
#     Scalable and accurate implementation of generalized mixed models
# using GDS files
#
# Copyright (C) 2019-2025    Xiuwen Zheng / AbbVie-ComputationalGenomics
# License: GPL-3
#


#######################################################################
# Internal R functions

.load_lib <- quote(
    suppressPackageStartupMessages(library("SAIGEgds", quietly=TRUE)))

.trait_list <- c("quantitative", "binary")

.load_skat <- function(verbose=TRUE)
{
    if (requireNamespace("SKAT", quietly=TRUE))
    {
        if (isTRUE(verbose))
            cat("    SKAT package loaded for Efficient Resampling\n")
    } else {
        stop("The 'SKAT' package should be installed to enable ",
            "Efficient Resampling.")
    }
}

# Internal model initialization
.init_nullmod <- function(modobj, ii, maf, mac, missing, spa.pval, ER.mac,
    var.ratio, geno.ploidy, Sigma_inv, chol_inv_X_Sigma,
    maxMAF=1, wbeta=double(), num_wbuf=0L, ultra_mac=10, collapse_method="max")
{
    # check
    if (!is.numeric(var.ratio) || anyNA(var.ratio))
        stop("Invalid variance ratio in the SAIGE model.")
    if (!is.null(Sigma_inv))
    {
        if (!is.matrix(Sigma_inv) && !inherits(Sigma_inv, "dsCMatrix"))
        {
            stop("Sigma_inv should be a dense matrix or ",
                "a sparse matrix 'dsCMatrix'.")
        }
    }
    if (!is.null(chol_inv_X_Sigma) && !is.matrix(chol_inv_X_Sigma))
        stop("chol_inv_X_Sigma should be a dense matrix.")

    if (!identical(class(modobj$obj.noK), "list"))
    {
        if (identical(class(modobj$obj.noK), "SA_NULL"))
        {
            warning("For more accurate model building, ",
                "the null model should be built using SAIGEgds>=v1.9.1!",
                call.=FALSE, immediate.=TRUE)
        } else
            stop("Unknown model object.")
    }

    # initialize the internal model parameters
    y <- unname(modobj$obj.noK$y)[ii]
    mu <- unname(modobj$fitted.values)[ii]
    X1 <- modobj$obj.noK$X1[ii,, drop=FALSE]
    t_X1 <- t(X1)
    V <- modobj$obj.noK$V[ii]
    n <- length(ii)  # Num of samples
    K <- ncol(X1)    # Num of fixed-effect coefficients
    cateMAC <- attr(var.ratio, "cateMAC")
    mobj <- list(
        trait = match(modobj$trait.type, .trait_list),
        maf = maf, mac = mac, missing = missing, spa.pval = spa.pval,
        ER.mac = ER.mac, geno.ploidy = geno.ploidy,
        tau = modobj$tau,
        y = y, mu = mu, y_mu = y - mu,
        mu2 = mu * (1 - mu),
        # K x n_samp (K << n_samp, more efficient)
        t_XXVX_inv = t(modobj$obj.noK$XXVX_inv[ii,, drop=FALSE]),
        XV = modobj$obj.noK$XV[, ii, drop=FALSE],  # K x n_samp
        # K x n_samp
        t_XVX_inv_XV = t(modobj$obj.noK$XXVX_inv[ii,, drop=FALSE] * V),
        t_X = t_X1,  # K x n_samp
        var.ratio = var.ratio,
        vr_sqrt = sqrt(var.ratio),
        vr_inv_sqrt = 1/sqrt(var.ratio),
        vr_cateMAC = if (isFALSE(cateMAC)) double() else cateMAC,
        Sigma_inv = Sigma_inv, chol_inv_X_Sigma = chol_inv_X_Sigma,
        # buffer
        buf_dosage = double(n),
        buf_coeff = double(2L*K),
        buf_adj_g = double(n),
        buf_index = integer(n),
        buf_B = double(n),
        buf_g_tilde = double(n),
        buf_X1 = double(K),
        buf_spa = double(2L*n)
    )

    if (is.na(mobj$trait))
        stop("Invalid 'modobj$trait.type'.")

    # additional design matrix
    if (modobj$trait.type == "binary")
    {
        mobj$XVX <- t(X1) %*% (X1 * mobj$mu2)  # a matrix: K x K
        mobj$S_a <- colSums(X1 * mobj$y_mu)    # a vector of size K
    } else if (modobj$trait.type == "quantitative")
    {
        mobj$XVX <- t(X1) %*% X1               # a matrix: K x K
        mobj$S_a <- colSums(X1 * mobj$y_mu)    # a vector of size K
    } else
        stop("Invalid 'modobj$trait.type': ", modobj$trait.type, ".")

    if (!is.null(Sigma_inv))
    {
        mobj$Si_X <- as.matrix(t(Sigma_inv %*% X1))
        mobj$XVX_inv_XV_X_Si_X <-
            as.matrix(t_X1 %*% t(mobj$Si_X) %*% mobj$t_XVX_inv_XV)
    }

    # additional for aggregate tests
    mobj$maxMAF <- as.double(maxMAF)
    mobj$buf_wbeta <- as.double(wbeta)
    mobj$num_unitsz <- num_wbuf
    mobj$buf_unitsz <- matrix(0.0, nrow=num_wbuf, ncol=7L)
    mobj$ultra_mac <- ultra_mac

    i <- match(collapse_method, c("max", "sum"))
    if (is.na(i)) stop("Internal error in 'collapse.method'.")
    mobj$collapse_method <- i

    # output
    mobj
}

.dsnode <- function(gdsfile, nm, aggregate=FALSE)
{
    if (nm == "")
    {
        if (!aggregate)
        {
            if (exist.gdsn(gdsfile, "genotype/data"))
            {
                nm <- "$dosage_alt2"
            } else {
                nm <- getOption("seqarray.node_ds", "annotation/format/DS")
                if (!exist.gdsn(gdsfile, nm))
                {
                    stop("Dosages should be stored in genotype or ",
                        "annotation/format/DS.")
                }
            }
        } else {
            nm <- "$dosage_sp2"
        }
    }
    nm
}

# check compression option
.check_compress <- function(res.compress)
{
    if (!is.character(res.compress))
        stop("'res.compress' should be character.")
    if (length(res.compress) != 1L)
        stop("'res.compress' should be character vector of length 1.")
    cp <- c("ZIP", "ZIP_RA", "LZMA", "LZMA_RA", "none")
    if (!(res.compress %in% cp))
    {
        stop("`res.compress` should be one of ",
            paste(paste(cp[-length(cp)], collapse=", "), "and", cp[length(cp)]),
            ".")
    }
}

# get the variance ratio estimate
.get_var_ratio <- function(modobj)
{
    # categorical MACs
    cateMAC <- modobj$use.cateMAC
    if (is.null(cateMAC)) cateMAC <- FALSE
    cateMAC <- .check_use_cateMAC(cateMAC)
    # variance ratio(s)
    if (is.numeric(modobj$var.ratio))
    {
        vr <- modobj$var.ratio[1L]
        if (!isFALSE(cateMAC))
            stop("'use.cateMAC' should be FALSE for a single variance ratio!")
    } else if (is.data.frame(modobj$var.ratio))
    {
        if (isFALSE(cateMAC))
        {
            vr <- mean(modobj$var.ratio$ratio, na.rm=TRUE)
        } else {
            if (!is.numeric(cateMAC))
            {
                stop("'use.cateMAC' should be a numeric vector for ",
                    "multiple variance ratios.")
            }
            x <- modobj$var.ratio$ratio
            d <- cut(modobj$var.ratio$mac, c(0, cateMAC, Inf),
                right=FALSE, dig.lab=15L)
            vr <- suppressWarnings(
                vapply(levels(d), function(s) mean(x[d==s]), 0))
            if (length(vr) != length(cateMAC)+1L)
            {
                stop("'use.cateMAC' should be a ", length(vr)-1L,
                    "-length vector!")
            }
            if (anyNA(vr))
            {
                print(vr)
                stop("variance ratios should not contain NaN.")
            }
        }
    } else
        stop("Invalid variance ratio in the model!")
    if (any(!is.finite(vr)))
        stop("The variance ratio should be a finite number.")
    attr(vr, "cateMAC") <- cateMAC
    vr
}

# save to R object file
.save_R_obj <- function(obj, res.compress, res.savefn, verbose)
{
    isfn <- !is.na(res.savefn) && res.savefn!=""
    if (isfn)
    {
        cm <- switch(res.compress, LZMA="xz", LZMA_RA="xz", ZIP="gzip",
            ZIP_RA="gzip", TRUE)
        if (verbose)
            .cat("Save to ", sQuote(res.savefn), " ...")
        if (grepl("\\.(rda|RData)$", res.savefn, ignore.case=TRUE))
        {
            .res <- obj
            save(.res, file=res.savefn, compress=cm)
        } else if (grepl("\\.rds$", res.savefn, ignore.case=TRUE))
        {
            saveRDS(obj, file=res.savefn, compress=cm)
        } else {
            stop("Unknown format of the output file, and it should be ",
                "RData, RDS or gds.")
        }
        if (verbose)
        {
            .cat(.crayon_underline(.tm()))
            .cat(.crayon_inverse("Done."))
        }
        invisible()
    } else {
        if (verbose)
        {
            .cat(.crayon_underline(.tm()))
            .cat(.crayon_inverse("Done."))
        }
        obj
    }
}



#######################################################################
# SAIGE single variant analysis
#

PVAL_METHOD_LEVELS <- c("Normal", "SPA", "ER")

.pval_method <- function(i)
{
    i <- suppressWarnings(as.integer(i))
    attr(i, "levels") <- PVAL_METHOD_LEVELS
    attr(i, "class") <- "factor"
    i
}


seqAssocGLMM_SPA <- function(gdsfile, modobj, maf=NaN, mac=10, missing=0.05,
    spa=TRUE, ER.mac=4.5, dsnode="", geno.ploidy=2L, res.savefn="",
    res.compress="ZIP", parallel=FALSE, load.balancing=TRUE,
    verbose.pval=c(0, 5e-10, 5e-8, 5e-6, 5e-4, 1), verbose=TRUE)
{
    stopifnot(inherits(gdsfile, "SeqVarGDSClass") | is.character(gdsfile))
    stopifnot(is.numeric(maf), length(maf)==1L)
    stopifnot(is.numeric(mac), length(mac)==1L)
    stopifnot(is.numeric(missing), length(missing)==1L)
    stopifnot(is.logical(spa), length(spa)==1L)
    stopifnot(is.numeric(ER.mac), length(ER.mac)==1L)
    stopifnot(is.character(dsnode), length(dsnode)==1L, !is.na(dsnode))
    stopifnot(is.numeric(geno.ploidy) | is.na(geno.ploidy),
        length(geno.ploidy)==1L)
    if (is.numeric(geno.ploidy) && !is.na(geno.ploidy))
        stopifnot(geno.ploidy >= 0L)
    stopifnot(is.character(res.savefn), length(res.savefn)==1L)
    .check_compress(res.compress)
    stopifnot(is.logical(load.balancing), length(load.balancing)==1L,
        !is.na(load.balancing))
    stopifnot(is.numeric(verbose.pval) | is.null(verbose.pval))
    stopifnot(is.logical(verbose), length(verbose)==1L)

    if (verbose)
    {
        cat(.crayon_inverse("SAIGE association analysis:\n"))
        .cat(.crayon_underline(.tm()))
    }

    # check model
    modobj <- .check_modobj(modobj, verbose)
    var.ratio <- .get_var_ratio(modobj)
    spa.pval <- if (isTRUE(spa)) NaN else -1
    if (is.na(ER.mac)) ER.mac <- 0

    # GDS file
    if (is.character(gdsfile))
    {
        if (verbose)
            .cat("    open ", sQuote(basename(gdsfile)))
        gdsfile <- seqOpen(gdsfile, allow.duplicate=TRUE)
        on.exit(seqClose(gdsfile))
    } else {
        # save the filter on GDS file
        seqFilterPush(gdsfile)
        on.exit(seqFilterPop(gdsfile))
    }

    # show warnings immediately
    saveopt <- options(warn=1L)
    on.exit(options(warn=saveopt$warn), add=TRUE)

    # determine the GDS node for dosages
    dsnode <- .dsnode(gdsfile, dsnode)
    using_gt <- dsnode=="$dosage_alt2"

    # check sample ID
    seqSetFilter(gdsfile, sample.id=modobj$sample.id, warn=FALSE, verbose=FALSE)
    sid <- seqGetData(gdsfile, "sample.id")
    if (length(sid) != length(modobj$sample.id))
        stop("Some of sample IDs are not available in the GDS file.")
    ii <- match(sid, modobj$sample.id)
    if (any(is.na(ii)))
        stop("Sample IDs do not match.")

    dm <- seqSummary(gdsfile, "genotype", verbose=FALSE)$seldim
    nVariant <- dm[3L]
    if (verbose)
    {
        .cat("    trait type: ", modobj$trait.type)
        .cat("    # of samples: ", .pretty(dm[2L]))
        .cat("    # of variants: ", .pretty(dm[3L]))
        if (!is.na(geno.ploidy) && (geno.ploidy>0L))
        {
            .cat("    MAF threshold: ", .pretty_gt_eq(maf))
            .cat("    MAC threshold: ", .pretty_gt_eq(mac))
            if (geno.ploidy != 2L)
                .cat("    ploidy: ", geno.ploidy)
        } else {
            cat("    not genotypic data\n")
        }
        .cat("    missing proportion threshold: ", .pretty_lt_eq(missing))
        v <- var.ratio
        if (isFALSE(attr(v, "cateMAC")))
        {
            .cat("    variance ratio for approximation: ", v)
        } else {
            cat("    variance ratios for approximation (MAC categories):\n")
            attr(v, "cateMAC") <- NULL
            print(v, width=1024L)
        }
        if (is.matrix(modobj$Sigma_inv))
            cat("    using a dense matrix of Sigma (covariance)\n")
        else if (inherits(modobj$Sigma_inv, "sparseMatrix"))
            cat("    using a sparse matrix of Sigma (covariance)\n")
        if (!isTRUE(spa) && (modobj$trait.type=="binary"))
            cat("    no adjustment for case-control imbalance\n")
    }

    if (dm[2L] <= 0) stop("No sample in the genotypic data set!")
    if (dm[3L] <= 0) stop("No variant in the genotypic data set!")

    # initialize the internal model parameters
    mobj <- .init_nullmod(modobj, ii, maf, mac, missing, spa.pval, ER.mac,
        var.ratio, geno.ploidy, modobj$Sigma_inv, modobj$chol_inv_X_Sigma)

    # load package(s)
    if (ER.mac >= mac) .load_skat(verbose)

    # update parallel object
    parallel <- SeqArray:::.McoreParallel(parallel)
    njobs <- SeqArray:::.NumParallel(parallel)
    is_fork <- SeqArray:::.IsForking(parallel)  # is forking or not?
    if (verbose)
        .cat("    # of processes: ", njobs)

    # initialize internally
    if (njobs<=1L || is_fork)
    {
        # forking, no need to distribute model parameters
        .Call(saige_score_test_init, mobj)
        initfun <- finalfun <- NULL
    } else {
        # pass the model parameters to each process
        if (verbose)
            cat("Distribute the model parameters to the", njobs, "processes\n")
        # initialize child processes internally
        initfun <- function(proc_id, mobj)
        {
            eval(.load_lib)
            .PkgEnv$modobj <- mobj
            .Call(saige_score_test_init, mobj)
        }
        # clear when exit
        finalfun <- function(proc_id, param)
        {
            .PkgEnv$modobj <- NULL
            remove(modobj, envir=.PkgEnv)
            gc(verbose=FALSE, reset=TRUE, full=TRUE)  # reset memory
            invisible()
        }
    }

    # reset memory GC to reduce memory usage (especially in forked processes)
    gc(verbose=FALSE, reset=TRUE, full=TRUE)

    # defined functions
    isfn <- !is.na(res.savefn) && res.savefn!=""
    isfn_gds <- isfn && grepl("\\.gds$", res.savefn, ignore.case=TRUE)
    if (!isfn_gds)
    {
        combine_fun <- "unlist"
        scan_fun <- function(f, dsnode, pverbose)
        {
            seqApply(f, dsnode, .cfunction("saige_score_test_pval"),
                as.is="list", parallel=FALSE, .list_dup=FALSE, .useraw=NA,
                .progress=SeqArray:::.process_verbose(pverbose))
        }
    } else {
        # output to a GDS file
        Append <- function(nm, val)
            append.gdsn(index.gdsn(outf, nm), val)
        combine_fun <- function(v)
        {
            Append("id", v$id)
            if (!is.na(geno.ploidy) && (geno.ploidy>0L))
            {
                Append("AF.alt", vapply(v$rv, `[`, 0, i=1L))
                Append("mac", vapply(v$rv, `[`, 0, i=2L))
            } else {
                Append("mean", vapply(v$rv, `[`, 0, i=1L))
                Append("nnzero", vapply(v$rv, `[`, 0, i=2L))
            }
            Append("num", vapply(v$rv, `[`, 0, i=3L))
            Append("beta", vapply(v$rv, `[`, 0, i=4L))
            Append("SE", vapply(v$rv, `[`, 0, i=5L))
            Append("pval", vapply(v$rv, `[`, 0, i=6L))
            Append("method", as.integer(vapply(v$rv, `[`, 0, i=7L)))
            if (modobj$trait.type == "binary")
            {
                Append("p.norm", vapply(v$rv, `[`, 0, i=8L))
                Append("converged", vapply(v$rv, `[`, 0, i=9L)==1L)
            }
            NULL
        }
        scan_fun <- function(f, dsnode, pverbose)
        {
            v<- seqApply(f, dsnode, .cfunction("saige_score_test_pval"),
                as.is="list", parallel=FALSE, .list_dup=FALSE, .useraw=NA,
                .progress=SeqArray:::.process_verbose(pverbose))
            i <- seqGetData(f, "$variant_index")
            x <- !vapply(v, is.null, FALSE)
            list(id=i[x], rv=v[x])
        }

        if (verbose)
            .cat("Save to ", sQuote(res.savefn), " ...")
        # create a GDS file
        cm <- res.compress[1L]
        outf <- createfn.gds(res.savefn)
        on.exit({ if (!is.null(outf)) closefn.gds(outf) }, add=TRUE)
        put.attr.gdsn(outf$root, "FileFormat", "SAIGE_OUTPUT")
        put.attr.gdsn(outf$root, "Version",
            paste0("SAIGEgds_", packageVersion("SAIGEgds")))

        # add function
        Add <- function(varnm, val)
            add.gdsn(outf, varnm, val, compress=cm, closezip=TRUE, replace=TRUE)
        # add sample IDs
        Add("sample.id", seqGetData(gdsfile, "sample.id"))
        add.gdsn(outf, "id", integer())
        if (!is.na(geno.ploidy) && (geno.ploidy>0L))
        {
            add.gdsn(outf, "AF.alt", double())
            add.gdsn(outf, "mac", if (using_gt) integer() else double())
        } else {
            add.gdsn(outf, "mean", double())
            add.gdsn(outf, "nnzero", integer())
        }
        add.gdsn(outf, "num", integer())
        add.gdsn(outf, "beta", double())
        add.gdsn(outf, "SE", double())
        add.gdsn(outf, "pval", double())
        n <- add.gdsn(outf, "method", integer())
        put.attr.gdsn(n, "R.class", "factor")
        put.attr.gdsn(n, "R.levels", PVAL_METHOD_LEVELS)
        if (modobj$trait.type == "binary")
        {
            add.gdsn(outf, "p.norm", double())
            add.gdsn(outf, "converged", logical())
        }
    }

    # scan variants
    rv <- seqParallel(parallel, gdsfile, FUN=scan_fun, split="by.variant",
        .combine=combine_fun, .initialize=initfun, .finalize=finalfun,
        .initparam=mobj, .balancing=load.balancing, .status_file=TRUE,
        .proc_time=verbose, dsnode=dsnode, pverbose=verbose)

    # output to a GDS file?
    if (!isfn_gds)
    {
        # if any maf/mac filter
        if (length(rv) != nVariant)
            stop("seqParallel() returns a vector of wrong length.")
        x <- vapply(rv, is.null, FALSE)
        if (any(x))
        {
            x <- !x
            seqSetFilter(gdsfile, variant.sel=x, action="intersect",
                verbose=FALSE)
            rv <- rv[x]
        }
        if (verbose)
        {
            cat("# of variants after filtering by ")
            if (!is.na(geno.ploidy) && (geno.ploidy>0L))
                cat("MAF, MAC and ")
            .cat("missing thresholds: ", .pretty(length(rv)))
        }

        # output
        ans <- data.frame(
            id  = seqGetData(gdsfile, "variant.id"),
            chr = seqGetData(gdsfile, "chromosome"),
            pos = seqGetData(gdsfile, "position"),
            stringsAsFactors = FALSE
        )
        # add RS IDs if exists
        if (exist.gdsn(gdsfile, "annotation/id"))
            ans$rs.id <- seqGetData(gdsfile, "annotation/id")
        ans$ref <- seqGetData(gdsfile, "$ref")
        ans$alt <- seqGetData(gdsfile, "$alt")
        if (!is.na(geno.ploidy) && (geno.ploidy>0L))
        {
            ans$AF.alt <- vapply(rv, `[`, 0, i=1L)
            ans$mac <- vapply(rv, `[`, 0, i=2L)
            if (using_gt) ans$mac <- as.integer(ans$mac)
        } else {
            ans$mean <- vapply(rv, `[`, 0, i=1L)
            ans$nnzero <- as.integer(vapply(rv, `[`, 0, i=2L))
        }
        ans$num  <- as.integer(vapply(rv, `[`, 0, i=3L))
        ans$beta <- vapply(rv, `[`, 0, i=4L)
        ans$SE   <- vapply(rv, `[`, 0, i=5L)
        ans$pval <- vapply(rv, `[`, 0, i=6L)
        ans$method <- .pval_method(vapply(rv, `[`, 0, i=7L))
        if (modobj$trait.type == "binary")
        {
            ans$p.norm <- vapply(rv, `[`, 0, i=8L)
            ans$converged <- vapply(rv, `[`, 0, i=9L)==1L
        }

        if (verbose && length(verbose.pval))
        {
            cat("P-value:")
            print(table(cut(ans$pval, breaks=unique(sort(verbose.pval)),
                        include.lowest=TRUE), exclude=NULL))
        }
        # save file?
        .save_R_obj(ans, res.compress, res.savefn, verbose)

    } else {
        # output to a GDS file
        # write data
        i <- read.gdsn(index.gdsn(outf, "id"))
        if (verbose)
        {
            cat("# of variants after filtering by ")
            if (!is.na(geno.ploidy) && (geno.ploidy>0L))
                cat("MAF, MAC and ")
            .cat("missing thresholds: ", .pretty(length(i)))
        }
        seqSetFilter(gdsfile, variant.sel=i, warn=FALSE, verbose=FALSE)
        nd_id <- .write_gds(outf, "id", gdsfile, "variant.id", cm)
        nd_chr <- .write_gds(outf, "chr", gdsfile, "chromosome", cm)
        moveto.gdsn(nd_chr, nd_id)
        nd_pos <- .write_gds(outf, "pos", gdsfile, "position", cm)
        moveto.gdsn(nd_pos, nd_chr)
        nd <- nd_pos
        # rs.id
        if (!is.null(index.gdsn(gdsfile, "annotation/id", silent=TRUE)))
        {
            nd <- .write_gds(outf, "rs.id", gdsfile, "annotation/id", cm)
            moveto.gdsn(nd, nd_pos)
        }
        # ref and alt alleles
        nd_ref <- Add("ref", seqGetData(gdsfile, "$ref"))
        moveto.gdsn(nd_ref, nd)
        nd_alt <- Add("alt", seqGetData(gdsfile, "$alt"))
        moveto.gdsn(nd_alt, nd_ref)
        # sub variables
        i <- order(i)
        for (nm in c("AF.alt", "mac", "mean", "nnzero", "num", "beta", "SE",
                "pval", "method", "p.norm", "converged"))
        {
            nd <- index.gdsn(outf, nm, silent=TRUE)
            if (!is.null(nd)) Add(nm, read.gdsn(nd)[i])
        }

        if (verbose && length(verbose.pval))
        {
            cat("P-value:")
            p <- read.gdsn(index.gdsn(outf, "pval"))
            print(table(cut(p, breaks=unique(sort(verbose.pval)),
                        include.lowest=TRUE), exclude=NULL))
        }

        # close the GDS file
        closefn.gds(outf); outf <- NULL
        cleanup.gds(res.savefn, verbose=FALSE)
        if (verbose)
        {
            .cat(.crayon_underline(.tm()))
            .cat(.crayon_inverse("Done."))
        }
        invisible()
    }
}

#######################################################################
#
# Package name: SAIGEgds
#
# Description:
#     Scalable and accurate implementation of generalized mixed models
# using GDS files
#
# Copyright (C) 2019-2026    Xiuwen Zheng / AbbVie-ComputationalGenomics
# License: GPL-3
#


#######################################################################
# Fit the null model
#

# check grm.mat
.check_grm_mat <- function(grm.mat, verbose)
{
    # check class
    if (inherits(grm.mat, "snpgdsGRMClass"))
    {
        m <- grm.mat$grm
        colnames(m) <- rownames(m) <- grm.mat$sample.id
        grm.mat <- m
    } else if (is.character(grm.mat))
    {
        if (length(grm.mat) != 1L)
            stop("'grm.mat' should be a file name.")
        if (verbose)
        {
            cat("Load a user-defined GRM:\n")
            .cat("    File: ", grm.mat)
        }
        if (grepl("\\.rds$", grm.mat, ignore.case=TRUE))
        {
            grm.mat <- readRDS(grm.mat)
        } else if (grepl("\\.(rda|RData)$", grm.mat, ignore.case=TRUE))
        {
            grm.mat <- get(load(grm.mat))
        } else
            stop("Unknown format of ", sQuote(grm.mat))
        if (verbose)
        {
            cat("    ")
            if (inherits(grm.mat, "sparseMatrix"))
                cat("Sparse matrix: ")
            else if (is.matrix(grm.mat))
                cat("Dense matrix: ")
            print(object.size(grm.mat))
        }
    }
    if (is.matrix(grm.mat) || inherits(grm.mat, "sparseMatrix"))
    {
        if (nrow(grm.mat) != ncol(grm.mat))
            stop("grm.mat should a numeric square matrix.")
        if ((is.matrix(grm.mat) && !is.double(grm.mat)) ||
                (inherits(grm.mat, "sparseMatrix") && !is.double(grm.mat@x)))
        {
            stop("grm.mat should be a numeric matrix.")
        }
        if (is.null(colnames(grm.mat)) || is.null(rownames(grm.mat)))
            stop("The row and column names of 'grm.mat' should be sample IDs.")
        if (!identical(colnames(grm.mat), rownames(grm.mat)))
            stop("The row and column names of 'grm.mat' should be the same.")
        if (anyDuplicated(colnames(grm.mat)))
            stop("The sample IDs in 'grm.mat' should be unique.")
    } else if (!is.null(grm.mat) && !isTRUE(grm.mat))
        stop("grm.mat should be NULL, a matrix, a file name or TRUE.")
    # check
    d <- diag(grm.mat)  # diag(TRUE)=1
    if (anyNA(d) || any(d <= 0))
        stop("The diagonal of GRM should be all positive.")
    # output
    grm.mat
}


# coerce to dgCMatrix (column-oriented sparse form)
.sp_to_dgCMatrix <- function(m)
{
    stopifnot(inherits(m, "sparseMatrix"))
    # if RsparseMatrix, convert to TsparseMatrix
    if (inherits(m, "RsparseMatrix"))
        m <- as(m, "TsparseMatrix")
    if (inherits(m, "TsparseMatrix"))
        m <- as(m, "CsparseMatrix")
    if (inherits(m, "dsCMatrix"))
        m <- as(m, "generalMatrix")
    if (!inherits(m, "dgCMatrix"))
        m <- as(m, "dgCMatrix")
    stopifnot(is(m, "dgCMatrix"))
    m
}


# get diagnoal of the full GRM
.get_grm_diag <- function() .Call(saige_get_grm_diag)


# calculate the variance ratio
.calcVR <- function(gdsfile, seed, fit0, glmm, obj.noK, param, verbose)
{
    var.ratio <- 1
    if (!is.null(gdsfile))
    {
        if (verbose)
        {
            .cat(.crayon_underline(.tm()))
            .cat(.crayon_inverse("Calculate the average ratio of variances:"))
        }
        set.seed(seed)
        var.ratio <- .Call(saige_calc_var_ratio, fit0, glmm, obj.noK, param)
    }
    glmm$var.ratio <- var.ratio
    glmm
}


# binary outcome
.fit_binary <- function(verbose, X.transform, phenovar, data, formula, param,
    tau.init, gdsfile, grm.mat, seed, n_var)
{
    if (verbose)
    {
        .cat("Binary outcome: ", phenovar)
        if (isTRUE(X.transform))
            y <- data$y
        else
            y <- data[[phenovar]]
        v <- table(y)
        n <- length(v) 
        v <- data.frame(v, as.numeric(prop.table(v)))
        v[, 1L] <- paste0("      ", v[, 1L])
        colnames(v) <- c(phenovar, "Number", "Proportion")
        rowmore <- nrow(v) > 10L
        if (rowmore) v <- v[1:10, ]
        print(v, row.names=FALSE)
        if (rowmore) cat("        ...\n")
        if (n > 2L)
            stop("The outcome variable has more than 2 categories!")
        else if (n <= 1L)
            stop("The outcome variable has only a unique value!")
    }

    # fit the null model
    fit0 <- glm(formula, data=data, family=binomial, offset=param$covoffset)
    if (verbose)
    {
        if (is.null(gdsfile) && is.null(grm.mat))
            cat("Fixed-effect coefficients:\n")
        else
            cat("Initial fixed-effect coefficients:\n")
        v <- as.data.frame(t(fit0$coefficients))
        rownames(v) <- "   "
        print(v, width=128L)
    }

    # design matrix
    X <- unname(model.matrix(fit0))
    attr(X, "assign") <- attr(X, "contrasts") <- NULL

    if (!is.null(gdsfile) || !is.null(grm.mat))
    {
        # initial tau
        tau <- fixtau <- c(0, 0)
        if (fit0$family$family %in% c("binomial", "poisson"))
            tau[1] <- fixtau[1] <- 1
        if (sum(tau.init[fixtau==0]) == 0)
            tau[fixtau==0] <- 0.1
        else
            tau[fixtau==0] <- tau.init[fixtau==0]
        # iterate
        glmm <- .Call(saige_fit_AI_PCG, fit0, X, tau, param)
    } else {
        # no random effect
        glmm <- list(
            coefficients = unname(fit0$coefficients),
            tau = c(1, 0),
            linear.predictors = unname(fit0$linear.predictors),
            fitted.values = unname(fit0$fitted.values),
            residuals = unname(fit0$residuals),
            converged = fit0$converged)
    }

    # use updated mu to set obj.noK (ScoreTest_NULL_Model)
    if (!is.null(param$Xmat)) X <- param$Xmat
    mu <- glmm$fitted.values
    V <- mu*(1-mu)
    XV <- t(X * V)
    XVX_inv <- solve(crossprod(X, X * V))
    XXVX_inv <- X %*% XVX_inv
    obj.noK <- list(y=unname(fit0$y), mu=mu, V=V, X1=X, XV=XV,
        XXVX_inv=XXVX_inv)
    glmm$obj.noK <- obj.noK

    if (!is.null(grm.mat))
    {
        # get the inverse of the Sigma matrix
        sigma <- grm.mat
        if (!is.null(gdsfile))
            diag(sigma) <- .get_grm_diag()  # for better approximation
        sigma <- glmm$tau[2L] * sigma
        diag(sigma) <- diag(sigma) + 1/V
        colnames(sigma) <- rownames(sigma) <- NULL
        m <- chol2inv(chol(sigma))
        if (inherits(m, "sparseMatrix")) m <- as(m, "symmetricMatrix")
        glmm$Sigma_inv <- obj.noK$Sigma_inv <- m
        # get the part of projection matrix
        X1 <- obj.noK$X1
        s_X1 <- solve(sigma, X1)
        m <- as.matrix(solve(crossprod(X1, s_X1)))
        m <- t(as.matrix(tcrossprod(chol(m), s_X1)))
        dimnames(m) <- NULL
        glmm$chol_inv_X_Sigma <- m
    } else {
        obj.noK$Sigma_inv <- FALSE
    }

    # calculate the variance ratio
    .calcVR(gdsfile, seed, fit0, glmm, obj.noK, param, verbose)
}


# quantitative outcome
.fit_quant <- function(verbose, X.transform, phenovar, data, inv.norm, formula,
    param, tau.init, gdsfile, grm.mat, seed, n_var)
{
    if (verbose)
    {
        .cat("Quantitative outcome: ", phenovar)
        if (isTRUE(X.transform))
            y <- data$y
        else
            y <- data[[phenovar]]
        v <- data.frame(mean=mean(y), sd=sd(y), min=min(y), max=max(y))
        rownames(v) <- "   "
        print(v)
    }

    # inverse normal transformation
    if (inv.norm == "residuals")
    {
        if (isTRUE(X.transform)) phenovar <- "y"
        fit0 <- glm(formula, data=data)
        resid.sd <- sd(fit0$residuals)
        new.y <- .rank_norm(fit0$residuals, s=resid.sd)
        data[[phenovar]] <- new.y
        if (verbose)
        {
            cat("Inverse normal transformation on the residuals with:\n")
            .cat("    standard deviation = ", resid.sd)
        }
    } else if (inv.norm == "quant")
    {
        if (isTRUE(X.transform)) phenovar <- "y"
        y <- data[[phenovar]]
        new.y <- .rank_norm(y)
        data[[phenovar]] <- new.y
        if (verbose)
            cat("Inverse normal transformation on the outcome variable\n")
    }

    # fit the null model
    fit0 <- glm(formula, data=data)
    if (verbose)
    {
        if (is.null(gdsfile) && is.null(grm.mat))
            cat("Fixed-effect coefficients:\n")
        else
            cat("Initial fixed-effect coefficients:\n")
        v <- as.data.frame(t(fit0$coefficients))
        rownames(v) <- "   "
        print(v, width=128L)
    }

    # design matrix
    X <- unname(model.matrix(fit0))
    attr(X, "assign") <- attr(X, "contrasts") <- NULL

    if (!is.null(gdsfile) || !is.null(grm.mat))
    {
        y <- fit0$y
        offset <- fit0$offset
        if (is.null(offset)) offset <- rep(0, length(y))
        eta <- fit0$linear.predictors
        mu <- fit0$fitted.values
        mu.eta <- fit0$family$mu.eta(eta)
        Y <- eta - offset + (y - mu)/mu.eta
        # initial tau
        tau <- tau.init
        if (sum(tau) == 0) tau <- c(0.5, 0.5)
        tau <- var(Y) * tau / sum(tau)
        # iterate
        glmm <- .Call(saige_fit_AI_PCG, fit0, X, tau, param)
    } else {
        # no random effect
        glmm <- list(
            coefficients = unname(fit0$coefficients),
            tau = c(var(fit0$residuals), 0),
            linear.predictors = unname(fit0$linear.predictors),
            fitted.values = unname(fit0$fitted.values),
            residuals = unname(fit0$residuals),
            converged = fit0$converged)
        if (verbose)
            .cat("Estimated tau: (", glmm$tau[1L], ", ", glmm$tau[2L], ")")
    }

    obj.noK <- list(y=unname(fit0$y), mu=glmm$fitted.values,
        V=rep(1, length(fit0$y)),
        X1=X, XV=t(X), XXVX_inv=X %*% solve(crossprod(X)))
    glmm$obj.noK <- obj.noK

    if (!is.null(grm.mat))
    {
        # get the inverse of the Sigma matrix
        sigma <- grm.mat
        if (!is.null(gdsfile))
            diag(sigma) <- .get_grm_diag()  # for better approximation
        sigma <- glmm$tau[2L] * sigma
        diag(sigma) <- diag(sigma) + glmm$tau[1L]
        colnames(sigma) <- rownames(sigma) <- NULL
        m <- chol2inv(chol(sigma))
        if (inherits(m, "sparseMatrix")) m <- as(m, "symmetricMatrix")
        glmm$Sigma_inv <- obj.noK$Sigma_inv <- m
        # get the part of projection matrix
        s_X <- solve(sigma, X)
        m <- as.matrix(solve(crossprod(X, s_X)))
        m <- t(as.matrix(tcrossprod(chol(m), s_X)))
        dimnames(m) <- NULL
        glmm$chol_inv_X_Sigma <- m
    } else {
        obj.noK$Sigma_inv <- FALSE
    }

    # calculate the variance ratio
    .calcVR(gdsfile, seed, fit0, glmm, obj.noK, param, verbose)
}


# check use.cateMAC
.cateMAC_default <- c(5.5, 10.5, 20.5)

.check_use_cateMAC <- function(use.cateMAC)
{
    err <- "'use.cateMAC' should be FALSE, TRUE or a numeric vector."
    if (is.logical(use.cateMAC))
    {
        if (length(use.cateMAC) != 1L) stop(err)
        if (is.na(use.cateMAC)) stop(err)
        if (use.cateMAC)
            use.cateMAC <- .cateMAC_default
    } else if (is.numeric(use.cateMAC))
    {
        if (!is.vector(use.cateMAC) || length(use.cateMAC)<=0L)
            stop(err)
        if (anyNA(use.cateMAC))
            stop("'use.cateMAC' should not contain NA or NaN.")
        if (is.unsorted(use.cateMAC, strictly=TRUE))
            stop("'use.cateMAC' should be strictly increasing.")
        if (use.cateMAC[1L] <= 0)
            stop("'use.cateMAC[1]' should be > 0.")
        if (!all(is.finite(use.cateMAC)))
            stop("The numeric values in 'use.cateMAC' should be finite.")
    }
    use.cateMAC
}

# show use.cateMAC
.show_use_cateMAC <- function(use.cateMAC, verbose)
{
    if (verbose)
    {
        cat("MAC categories:")
        if (is.logical(use.cateMAC))
        {
            if (isFALSE(use.cateMAC))
                cat(" none\n")
            else
                .cat(" ", use.cateMAC)
        } else {
            x1 <- c(0, use.cateMAC)
            x2 <- c(use.cateMAC, Inf)
            s <- paste0("[", x1, ",", x2, ")")
            .cat("\n    ", paste(s, collapse=", "))
        }
    }
    invisible()
}

# simulated genotype packed RAW matrix (mac_low <= MAC < mac_high)
.simu_geno <- function(nsamp, nsnp, mac_low, mac_high)
{
    lv <- ceiling(mac_low)
    hv <- floor(mac_high)
    if (hv == mac_high) hv <- hv - 1L
    if (lv <= hv)
    {
        gmat <- matrix(as.raw(0L), nrow=ceiling(nsamp/4L), ncol=nsnp)
        mac_int <- seq.int(lv, hv)
        for (i in seq_len(nsnp))
        {
            if (length(mac_int) > 1L)
                mac <- sample(mac_int, 1L)
            else
                mac <- mac_int
            j <- sample.int(2L*nsamp, mac)
            # set genotypes according to j
            .Call(saige_set_geno2b_raw, gmat, j, i)
        }
        gmat
    } else {
        stop(sprintf("No integer MAC in [%g, %g).", mac_low, mac_high))
    }
}

# get a sparse form of genotypes
.get_sparse_geno <- function(gdsfile, nproc, verbose)
{
    # check
    if (is.character(gdsfile))
    {
        stopifnot(length(gdsfile)==1L)
        if (isTRUE(verbose))
            .cat("Open ", sQuote(basename(gdsfile)))
        gdsfile <- seqOpen(gdsfile, allow.duplicate=TRUE)
        on.exit(seqClose(gdsfile))
    } else {
        stopifnot(inherits(gdsfile, "SeqVarGDSClass"))
    }
    # integer genotypes or numeric dosages
    if (exist.gdsn(gdsfile, "genotype/data"))
    {
        varnm <- "$dosage_alt2"
    } else if (exist.gdsn(gdsfile, "annotation/format/DS/data"))
    {
        varnm <- "annotation/format/DS"
        if (verbose) cat("    using 'annotation/format/DS'\n")
    } else {
        stop("'genotype' and 'annotation/format/DS' are not available.")
    }
    # initialize
    dm <- seqSummary(gdsfile, "genotype", verbose=FALSE)$seldim
    nsamp <- dm[2L]; nvar  <- dm[3L]
    remove(dm)
    # process
    stopifnot(is.numeric(nproc))
    if (nproc > 1L)
    {
        # parallel loading
        cl <- parallel::makeCluster(nproc)
        on.exit(parallel::stopCluster(cl), add=TRUE)
        bs <- ceiling(nvar/100L)  # block size
        seqParallel(cl, gdsfile, FUN=function(gds, varnm)
        {
            gc(FALSE, reset=TRUE)
            seqApply(gds, varnm, .cfunction("saige_get_sparse"),
                as.is="list", .useraw=TRUE, .list_dup=FALSE)
        },
        .initialize=function(proc_id, nsamp)
        {
            # internal buffer
            buf_b256 <- integer((ceiling(nsamp/256L) + 1L)*3L + 1L)
            buf_b1 <- raw(3L*ceiling(nsamp/256L)*257L)
            assign("buf_b256", buf_b256, envir=.PkgEnv)
            assign("buf_b1", buf_b1, envir=.PkgEnv)
            .Call(saige_init_sparse, nsamp, buf_b256, buf_b1)
        },
        .finalize=function(proc_id, nsamp)
        {
            .PkgEnv$buf_b256 <- .PkgEnv$buf_b1 <- NULL
            remove("buf_b256", "buf_b1", envir=.PkgEnv)
        }, .initparam=nsamp, .balancing=TRUE, .bl_size=bs, .bl_progress=verbose,
        varnm=varnm)
    } else {
        # internal buffer
        buf_b256 <- integer((ceiling(nsamp/256L) + 1L)*3L + 1L)
        buf_b1 <- raw(3L*ceiling(nsamp/256L)*257L)
        .Call(saige_init_sparse, nsamp, buf_b256, buf_b1)
        # apply
        seqApply(gdsfile, varnm, .cfunction("saige_get_sparse"),
            as.is="list", .useraw=TRUE, .list_dup=FALSE, .progress=verbose)
    }
}

# get information from the output of .get_sparse_geno()
.get_sparse_smallblock_len <- function(x)
{
    stopifnot(is.raw(x))
    .Call(saige_get_sparse_info, x)
}


# fit the null model
seqFitNullGLMM_SPA <- function(formula, data, gdsfile=NULL, grm.mat=NULL,
    trait.type=c("binary", "quantitative"), sample.col="sample.id", maf=0.01,
    missing.rate=0.01, max.num.snp=1000000L, variant.id=NULL,
    variant.id.varratio=NULL, nsnp.sub.random=2000L, rel.cutoff=0.125,
    inv.norm=c("residuals", "quant", "none"), use.cateMAC=FALSE,
    cateMAC.inc.maf=TRUE, cateMAC.simu=TRUE, use.offset=FALSE, X.transform=TRUE,
    tol=0.02, maxiter=20L, nrun=30L, tolPCG=1e-5, maxiterPCG=500L,
    num.marker=30L, tau.init=c(0,0), traceCVcutoff=0.0025, ratioCVcutoff=0.001,
    geno.sparse=TRUE, num.thread=1L, model.savefn="", seed=200L,
    fork.loading, parallel.loading=FALSE, verbose=TRUE)
{
    # check
    stopifnot(inherits(formula, "formula"))
    stopifnot(is.data.frame(data))
    stopifnot(is.null(gdsfile) || inherits(gdsfile, "SeqVarGDSClass") ||
            is.character(gdsfile))
    trait.type <- match.arg(trait.type)
    stopifnot(is.character(sample.col), length(sample.col)==1L,
        !is.na(sample.col))
    stopifnot(is.numeric(maf), length(maf)==1L)
    stopifnot(is.numeric(missing.rate), length(missing.rate)==1L)
    stopifnot(is.numeric(max.num.snp), length(max.num.snp)==1L)
    stopifnot(is.null(variant.id) | is.vector(variant.id))
    stopifnot(is.null(variant.id.varratio) | is.vector(variant.id.varratio))
    stopifnot(is.numeric(nsnp.sub.random), length(nsnp.sub.random)==1L,
        nsnp.sub.random>=0L)
    stopifnot(is.numeric(rel.cutoff), length(rel.cutoff)==1L)
    if (is.na(rel.cutoff)) rel.cutoff <- -Inf
    if (is.logical(inv.norm))
        inv.norm <- if (isTRUE(inv.norm)) "residuals" else "none"
    inv.norm <- match.arg(inv.norm)
    use.cateMAC <- .check_use_cateMAC(use.cateMAC)
    if (is.logical(cateMAC.inc.maf))
    {
        stopifnot(length(cateMAC.inc.maf) == 1L)
    } else if (is.numeric(cateMAC.inc.maf))
    {
        stopifnot(is.vector(cateMAC.inc.maf))
        if (anyNA(cateMAC.inc.maf))
            stop("'cateMAC.inc.maf' should not include NA/NaN.")
        if (any(cateMAC.inc.maf<=0 | cateMAC.inc.maf>=1))
            stop("'cateMAC.inc.maf' should be between 0 and 1.")
    } else {
        stop("'cateMAC.inc.maf' should be FALSE, TRUE or ",
            "a numeric vector for MAF.")
    }
    stopifnot(is.logical(cateMAC.simu), length(cateMAC.simu)==1L)
    stopifnot(is.logical(use.offset), length(use.offset)==1L)
    stopifnot(is.logical(X.transform), length(X.transform)==1L)
    stopifnot(is.numeric(tol), length(tol)==1L)
    stopifnot(is.numeric(maxiter), length(maxiter)==1L)
    stopifnot(is.numeric(nrun), length(nrun)==1L)
    stopifnot(is.numeric(tolPCG), length(tolPCG)==1L)
    stopifnot(is.numeric(maxiterPCG), length(maxiterPCG)==1L)
    stopifnot(is.numeric(num.marker), length(num.marker)==1L)
    stopifnot(is.numeric(tau.init), length(tau.init)==2L)
    stopifnot(is.numeric(traceCVcutoff), length(traceCVcutoff)==1L)
    stopifnot(is.numeric(ratioCVcutoff), length(ratioCVcutoff)==1L)
    stopifnot(is.logical(geno.sparse), length(geno.sparse)==1L)
    stopifnot(is.numeric(num.thread), length(num.thread)==1L)
    stopifnot(is.character(model.savefn), length(model.savefn)==1L)
    stopifnot(is.numeric(seed), length(seed)==1L, is.finite(seed))
    stopifnot(is.logical(parallel.loading), length(parallel.loading)==1L)
    stopifnot(is.logical(verbose), length(verbose)==1L)
    if (!missing(fork.loading))
    {
        warning("'fork.loading' is deprecated, ",
            "please use 'parallel.loading' instead.")
    }
    if (verbose)
    {
        .cat(.crayon_inverse("SAIGE association analysis:"))
        .cat(.crayon_underline(.tm()))
    }

    # check GRM matrix if specified
    if (isFALSE(grm.mat)) grm.mat <- NULL
    noRE <- is.null(gdsfile) && is.null(grm.mat)  # no random effect
    grm.mat <- .check_grm_mat(grm.mat, verbose)

    # GDS file
    if (is.character(gdsfile))
    {
        if (verbose)
            .cat("Open ", sQuote(gdsfile))
        gdsfile <- seqOpen(gdsfile, allow.duplicate=TRUE)
        on.exit(seqClose(gdsfile))
    } else if (!is.null(gdsfile))
    {
        # save the filter on GDS file
        seqFilterPush(gdsfile)
        on.exit(seqFilterPop(gdsfile))
    }

    # show warnings immediately
    saveopt <- options(warn=1L)
    on.exit(options(warn=saveopt$warn), add=TRUE)
    if (!is.null(seed)) set.seed(seed)

    # variables in the formula
    s <- as.character(formula)
    formula_str <- paste(s[2L], s[1L], s[3L])
    vars <- all.vars(formula)
    phenovar <- all.vars(formula)[1L]
    y <- data[[phenovar]]
    if (is.null(y))
        stop("There is no '", phenovar, "' in the input data frame.")
    if (!is.factor(y) && !is.numeric(y) && !is.logical(y))
        stop("The response variable should be numeric or a factor.")

    # check sample id
    if (sample.col %in% vars)
        stop(sprintf("'%s' should not be in the formula.", sample.col))
    if (!(sample.col %in% colnames(data)))
    {
        stop(sprintf("'%s' should be one of the columns in 'data'.",
                sample.col))
    }
    if (is.factor(data[[sample.col]]))
        stop(sprintf("'%s' should not be a factor variable.", sample.col))
    if (any(is.na(data[[sample.col]])))
        stop(sprintf("'%s' should not have any missing value.", sample.col))
    if (anyDuplicated(data[[sample.col]]))
        stop(sprintf("'%s' in data should be unique.", sample.col))

    # remove missing values
    data <- data[, c(sample.col, vars)]
    data <- na.omit(data)
    data <- droplevels(data)
    sid <- NULL
    if (!is.null(gdsfile))
    {
        seqResetFilter(gdsfile, sample=TRUE, verbose=FALSE)
        sid <- seqGetData(gdsfile, "sample.id")
    }
    if (!is.null(grm.mat) && !isTRUE(grm.mat))
    {
        if (is.null(sid))
            sid <- colnames(grm.mat)
        else
            sid <- intersect(sid, colnames(grm.mat))
        if (length(sid) <= 0L)
            stop("'gdsfile' and 'grm.mat' should have shared sample IDs.")
    } else if (is.null(sid))
        sid <- data[[sample.col]]
    i <- match(sid, data[[sample.col]])
    i <- i[!is.na(i)]
    data <- data[i, ]
    if (nrow(data) <= 0L)
        stop("No common sample.id between 'data' and the GDS file.")
    if (!is.null(gdsfile))
        seqSetFilter(gdsfile, sample.id=data[[sample.col]], verbose=FALSE)
    sid <- data[[sample.col]]

    if (!is.null(gdsfile))
    {
        # use gds genotype file
        if (is.null(variant.id))
            seqResetFilter(gdsfile, sample=FALSE, variant=TRUE, verbose=FALSE)
        else
            seqSetFilter(gdsfile, variant.id=variant.id, verbose=verbose)

        # filters of maf, mac, missing.rate
        if (verbose)
            cat("Filtering variants:\n")
        v <- seqGetAF_AC_Missing(gdsfile, minor=TRUE, parallel=num.thread,
            verbose=verbose)
        sel <- (v$miss <= missing.rate) & (v$ac > 0)
        sel[is.na(sel)] <- FALSE
        s1 <- sel & (v$af >= maf)  # variants used in GRM

        # if using the list of variant IDs to estimate variance ratio
        if (!is.null(variant.id.varratio))
        {
            seqFilterPush(gdsfile)
            seqSetFilter(gdsfile, variant.id=variant.id.varratio,
                verbose=FALSE)
            v <- seqGetAF_AC_Missing(gdsfile, minor=TRUE, parallel=num.thread,
                verbose=FALSE)
            sel <- (v$miss <= missing.rate) & (v$ac > 0)
            sel[is.na(sel)] <- FALSE
            if (verbose)
            {
                .cat("# of variants specified in ",
                    "the variance ratio estimation: ", nrow(v))
            }
        }

        # need random markers in variance ratio estimation
        set.seed(seed)
        if (isFALSE(use.cateMAC))
        {
            if (verbose)
                cat("MAC category for estimating variance ratio:\n")
            last <- 20  # mac threshold used in SAIGE
            mac_cat <- Inf
        } else {
            # using MAC categories
            stopifnot(is.numeric(use.cateMAC), is.vector(use.cateMAC))
            if (verbose)
                cat("MAC categories for estimating variance ratios:\n")
            if (isTRUE(cateMAC.inc.maf))
                cateMAC.inc.maf <- ifelse(is.finite(maf), maf, numeric())
            if (is.numeric(cateMAC.inc.maf))
            {
                for (m in sort(cateMAC.inc.maf))
                {
                    a <- 2L * length(sid) * m
                    if (all(abs(a-use.cateMAC) > 1L) && (a > max(use.cateMAC)))
                    {
                        use.cateMAC <- sort(c(use.cateMAC, a))
                        if (verbose)
                        {
                            .cat("    (including a cut point ", a,
                                " according to MAF=", m, ")")
                        }
                    }
                }
            }
            last <- .Machine$double.eps  # so that > 0
            mac_cat <- c(use.cateMAC, Inf)
        }
        rand.packed.geno <- vector("list", length(mac_cat))
        rand.packed.geno.vid <- NULL
        for (k in seq_along(mac_cat))
        {
            mac <- mac_cat[k]
            s2 <- sel & (last <= v$ac) & (v$ac < mac)
            ii <- which(s2)
            if (length(ii) < num.marker)
            {
                if (!isFALSE(use.cateMAC) && isTRUE(cateMAC.simu))
                {
                    if (verbose)
                    {
                        cat(sprintf("    MAC%s, %g):\t",
                                if (last > .Machine$double.eps)
                                    paste0("[", last) else "(0",
                                mac))
                    }
                    n <- 5L * num.marker
                    rand.packed.geno[[k]] <-
                        .simu_geno(length(sid), n, last, mac)
                    rand.packed.geno.vid <- c(rand.packed.geno.vid,
                        paste0("simu", k, "_g", seq_len(n)))
                    if (verbose)
                        cat(sprintf("%d+ simulated variants\n", num.marker))
                } else {
                    stop(sprintf("Less variants (n=%d) than %d in MAC[%g, %g)",
                            length(ii), num.marker, last, mac),
                        ", consider using simulated genotypes ",
                        "via 'cateMAC.simu=TRUE'.")
                }
            } else {
                n <- length(ii)
                if (length(ii) >= num.marker*5L)
                {
                    ii <- sample(ii, num.marker*5L)
                } else {
                    ii <- sample(ii, length(ii))
                }
                if (verbose)
                {
                    if (last > .Machine$double.eps)
                        a <- paste0("[", last)
                    else
                        a <- "(0"
                    if (n > num.marker)
                    {
                        .cat(sprintf(
                            "    MAC%s, %g):\t%d+ randomly from %s variants",
                            a, mac, num.marker, .pretty(n)))
                    } else {
                        .cat(sprintf("    MAC%s, %g):\t%d variants", a, mac, n))
                    }
                }
                seqFilterPush(gdsfile)
                seqSetFilter(gdsfile, variant.sel=ii, action="intersect",
                    verbose=FALSE)
                i <- match(ii, sort(ii))
                rand.packed.geno[[k]] <-
                    seqGet2bGeno(gdsfile, verbose=FALSE)[,i]
                rand.packed.geno.vid <- c(rand.packed.geno.vid,
                    seqGetData(gdsfile, "variant.id")[i])
                seqFilterPop(gdsfile)
            }
            last <- mac
        }

        # SNP markers used in GRM
        if (!is.null(variant.id.varratio))
            seqFilterPop(gdsfile)
        seqSetFilter(gdsfile, variant.sel=s1, action="intersect", verbose=FALSE)
        dm <- seqSummary(gdsfile, "genotype", verbose=FALSE)$seldim
        n_samp <- dm[2L]
        n_var  <- n <- dm[3L]
        if (max.num.snp>0L && n_var>max.num.snp)
        {
            set.seed(seed)
            v <- seqGetData(gdsfile, "$variant_index")
            seqSetFilter(gdsfile, variant.sel=sample(v, max.num.snp),
                warn=FALSE, verbose=FALSE)
            n_var <- as.integer(max.num.snp)
        }

        if (verbose)
        {
            .cat("Fit the null model: ", format(formula),
                ifelse(noRE, "", " + var(GRM)"))
            .cat("    # of samples: ", .pretty(n_samp))
            cat("    # of variants in GRM:", .pretty(n_var))
            if (n > max.num.snp)
                cat(" (randomly selected from ", .pretty(n), ")", sep="")
            cat("\n")
            .cat("    MAF threshold for GRM: >= ", maf)
        }
    } else {
        n_samp <- length(sid)
        n_var  <- NA_integer_
        rand.packed.geno <- NULL
        if (verbose)
        {
            .cat("Fit the null model: ", format(formula),
                ifelse(noRE, "", "+ var(GRM)"))
            .cat("    # of samples: ", .pretty(n_samp))
        }
    }

    # set the number of internal threads
    if (is.na(num.thread) || num.thread < 1L)
        num.thread <- 1L
    .Call(saige_set_numthread, num.thread)
    if (verbose)
    {
        .cat("    using ", num.thread, " thread",
            if (num.thread>1L) "s" else "")
    }

    # rearrange grm.mat if needed
    if (isTRUE(grm.mat))
    {
        # need a sparse GRM from the full GRM
        # correct nsnp.sub.random
        if (nsnp.sub.random > 0L)
        {
            nsnp.sub.random <- as.integer(floor(nsnp.sub.random/4) * 4L)
            if (nsnp.sub.random < 1L) nsnp.sub.random <- 4L
        }
        if (verbose)
        {
            cat("Calculating the sparse GRM:\n")
            .cat("    using relatedness threshold: ", rel.cutoff)
        }
        # calculate sparse GRM
        grm.mat <- .fit_calc_sp_grm(gdsfile, nsnp.sub.random, maf,
            missing.rate, rel.cutoff, num.thread, FALSE, FALSE, verbose)
        gc(verbose=FALSE, reset=TRUE, full=TRUE)  # reduce memory usage
        if (verbose) cat("Done (sparse GRM)\n")
    }
    if (!is.null(grm.mat) && !identical(sid, colnames(grm.mat)))
    {
        i <- match(sid, colnames(grm.mat))
        grm.mat <- grm.mat[i, i]
    }

    X <- model.matrix(formula, data)
    if (NCOL(X) <= 1L) use.offset <- X.transform <- FALSE

    # transform to avoid multi-collinearity and improve numeric stability
    if (isTRUE(X.transform))
    {
        if (verbose)
            cat("Transform on the design matrix with QR decomposition:\n")
        frm <- model.frame(formula, data)
        y <- model.response(frm, type="any")
        # check multi-collinearity
        m <- lm(y ~ X - 1)
        i_na <- which(is.na(m$coefficients))
        if (length(i_na) > 0L)
        {
            X <- X[, -i_na]
            if (verbose)
            {
                .cat("    exclude ", length(i_na), " covariates (",
                    paste(colnames(X)[i_na], collapse=", "),
                    ") to avoid multi collinearity.")
            }
        }
        X_name <- colnames(X)
        Xqr <- qr(X)  # QR decomposition
        X_new <- qr.Q(Xqr) * sqrt(nrow(X))
        X_qrr <- qr.R(Xqr)
        data <- data.frame(cbind(y, X_new))
        nm <- paste0("x_", seq_len(ncol(X_new))-1L)
        colnames(data) <- c("y", nm)
        formula <- as.formula(paste("y ~", paste(nm, collapse=" + "), "-1"))
        if (verbose)
            .cat("    new formula: ", format(formula))
    }

    # estimate the fixed effect coefficients or not
    covoffset <- Xmat <- NULL
    if (isTRUE(use.offset))
    {
        if (verbose)
        {
            cat("    using covariate offset instead of estimating",
                "each fixed effect coefficient\n")
        }
        Xmat <- model.matrix(formula, data=data)
        if (trait.type == "binary")
        {
            mod <- glm(formula, data=data, family=binomial)
        } else {
            mod <- glm(formula, data=data, family=gaussian)
        }
        covoffset <- Xmat[, -1L, drop=F] %*%  mod$coefficients[-1L]
        formula <- as.formula("y ~ 1")
    }	    

    # clear the internal GRM matrix
    .Call(saige_init_fit_grm)
    # internal buffer for diagonal of GRM
    buf_sigma_diag <- double(n_samp)

    # load SNP genotypes
    if (!is.null(gdsfile))
    {
        if (verbose)
        {
            cat("Use dense genetic relationship matrix in the file:\n")
            .cat("    ", gdsfile$filename)
            cat("Loading SNP genotypes from the GDS file:\n")
        }
        nproc <- 1L
        if (isTRUE(parallel.loading)) nproc <- num.thread
        if (isTRUE(geno.sparse))
        {
            # sparse genotypes
            packed.geno <- .get_sparse_geno(gdsfile, nproc, verbose)
        } else {
            # 2-bit packed genotypes
            packed.geno <- seqGet2bGeno(gdsfile, parallel=nproc,
                verbose=verbose)
        }
        if (verbose)
        {
            .cat("    using ",
                .pretty_size(as.double(object.size(packed.geno))),
                " (stored in a ",
                ifelse(isTRUE(geno.sparse), "sparse", "dense"), " form)")
        }

        # initialize internal variables and buffers
        buf_std_geno <- double(4L*n_var)
        buf_crossprod <- matrix(0.0, nrow=n_samp, ncol=num.thread)
        if (isTRUE(geno.sparse))
        {
            .Call(saige_store_sp_geno, packed.geno, rand.packed.geno,
                n_samp, buf_std_geno, buf_sigma_diag, buf_crossprod)
        } else {
            .Call(saige_store_2b_geno, packed.geno, rand.packed.geno,
                n_samp, buf_std_geno, buf_sigma_diag, buf_crossprod)
        }
    }

    # if use GRM matrix
    if (is.matrix(grm.mat))
    {
        # set dense GRM
        .Call(saige_store_dense_grm, n_samp, grm.mat, buf_sigma_diag)
        if (verbose)
        {
            cat("User-defined")
            if (!is.null(gdsfile)) cat(" approximate")
            cat(" genetic relationship matrix:\n")
            cat(sprintf("    %d x %d (dense matrix)\n", n_samp, n_samp))
        }
    } else if (inherits(grm.mat, "sparseMatrix"))
    {
        # column-oriented sparse form
        grm.mat <- .sp_to_dgCMatrix(grm.mat)
        # set sparse matrix
        .Call(saige_store_sparse_grm, n_samp, grm.mat, buf_sigma_diag)
        if (verbose)
        {
            a <- nnzero(grm.mat)
            s <- sprintf("%.3f%%", a/prod(dim(grm.mat))*100)
            if (s=="0.000%") s <- "<0.001%"
            if (!is.null(gdsfile)) cat("Approximate u") else cat("U")
            cat("ser-defined sparse genetic relationship matrix:\n")
            cat(sprintf("    %d x %d, # of nonzero: %d (%s)\n",
                    n_samp, n_samp, a, s))
        }
    } else if (is.null(grm.mat) && is.null(gdsfile) && verbose)
        cat("Assuming independent outcomes\n")

    # parameters for fitting the model
    param <- list(
        trait = match(trait.type, .trait_list),
        covoffset = covoffset, Xmat = Xmat,
        num.thread = num.thread, seed = seed,
        tol = tol, tolPCG = tolPCG,
        maxiter = maxiter, maxiterPCG = maxiterPCG, no_iteration = FALSE,
        nrun = nrun, num.marker = num.marker,
        traceCVcutoff = traceCVcutoff, ratioCVcutoff = ratioCVcutoff,
        verbose = verbose,
        indent = ""
    )

    tau.init[is.na(tau.init)] <- 0
    tau.init[tau.init < 0] <- 0

    # fit the model
    if (trait.type == "binary")
    {
        # logistic regression
        glmm <- .fit_binary(verbose, X.transform, phenovar, data, formula,
            param, tau.init, gdsfile, grm.mat, seed, n_var)
    } else if (trait.type == "quantitative")
    {
        # linear regression
        glmm <- .fit_quant(verbose, X.transform, phenovar, data, inv.norm,
            formula, param, tau.init, gdsfile, grm.mat, seed, n_var)
    } else
        stop("Invalid 'trait.type'.")

    glmm <- c(list(formula=formula_str), glmm)
    glmm$use.cateMAC <- use.cateMAC
    if (is.data.frame(glmm$var.ratio))
    {
        glmm$var.ratio$id <- rand.packed.geno.vid[glmm$var.ratio$id]
        if (verbose)
        {
            CV <- function(x) sd(x)/(mean(x)*length(x))
            if (isFALSE(use.cateMAC))
            {
                m <- mean(glmm$var.ratio$ratio)
                s <- sd(glmm$var.ratio$ratio)
                cv <- CV(glmm$var.ratio$ratio)
                .cat("    ratio avg: ", m, ", sd: ", s, ", CV: ", cv)
            } else {
                x <- glmm$var.ratio$ratio
                d <- cut(glmm$var.ratio$mac, c(0, use.cateMAC, Inf),
                    right=FALSE, dig.lab=15L)
                m <- suppressWarnings(data.frame(
                    n  = vapply(levels(d), function(s) sum(d==s), 0L),
                    mean = vapply(levels(d), function(s) mean(x[d==s]), 0),
                    sd = vapply(levels(d), function(s) sd(x[d==s]), 0),
                    CV = vapply(levels(d), function(s) CV(x[d==s]), 0),
                    stringsAsFactors=FALSE
                ))
                rownames(m) <- paste("   ", rownames(m))
                cat("    MAC categories:\n")
                print(m)
            }
        }
    }

    # tweak the result
    if (!isTRUE(X.transform) || isTRUE(use.offset))
    {
        if (isTRUE(use.offset))
            names(glmm$coefficients) <- "(Offset)"
        else
            names(glmm$coefficients) <- colnames(glmm$obj.noK$X1)
    } else {
        coef <- solve(X_qrr, glmm$coefficients * sqrt(nrow(data)))
        names(coef) <- X_name
        glmm$coefficients <- coef
    }
    names(glmm$tau) <- c("Sigma_E", "Sigma_G")
    glmm$trait.type <- trait.type
    if (!is.null(gdsfile))
    {
        glmm$sample.id <- seqGetData(gdsfile, "sample.id")
        glmm$variant.id <- seqGetData(gdsfile, "variant.id")
    } else {
        glmm$sample.id <- sid
        glmm$variant.id <- NULL
    }
    class(glmm) <- "ClassSAIGE_NullModel"

    if (!is.na(model.savefn) && model.savefn!="")
    {
        .cat("Save the model to ", sQuote(model.savefn))
        if (grepl("\\.(rda|RData)$", model.savefn, ignore.case=TRUE))
        {
            .glmm <- glmm
            save(.glmm, file=model.savefn)
        } else if (grepl("\\.rds$", model.savefn, ignore.case=TRUE))
        {
            saveRDS(glmm, file=model.savefn)
        } else {
            stop("Unknown format of the output file, ",
                "and it should be RData or RDS.")
        }
    }
    if (verbose)
    {
        .cat(.crayon_underline(.tm()))
        .cat(.crayon_inverse("Done."))
    }

    if (!is.na(model.savefn) && model.savefn!="")
        return(invisible(glmm))
    else
        return(glmm)
}

#######################################################################
#
# Package name: SAIGEgds
#
# Description:
#     Scalable and accurate implementation of generalized mixed models
# using GDS files
#
# Copyright (C) 2022-2024    Xiuwen Zheng / AbbVie-ComputationalGenomics
# License: GPL-3
#


# Select LD-pruned variants for genetic relationship matrix
seqFitLDpruning <- function(gdsfile, sample.id=NULL, variant.id=NULL,
    ld.threshold=0.1, maf=0.01, missing.rate=0.005, autosome.only=TRUE,
    use.cateMAC=TRUE, num.marker=100L, num.total=100000L, save.gdsfn=NULL,
    seed=200L, parallel=FALSE, parallel.multi=NULL, verbose=TRUE)
{
    stopifnot(inherits(gdsfile, "SeqVarGDSClass") | is.character(gdsfile))
    stopifnot(is.numeric(ld.threshold), length(ld.threshold)==1L)
    stopifnot(is.numeric(maf), length(maf)==1L)
    stopifnot(is.numeric(missing.rate), length(missing.rate)==1L)
    stopifnot(is.logical(autosome.only), length(autosome.only)==1L)
    use.cateMAC <- .check_use_cateMAC(use.cateMAC)
    stopifnot(is.numeric(num.marker), length(num.marker)==1L, num.marker>0L)
    stopifnot(is.numeric(num.total), length(num.total)==1L, num.total>0L)
    if (is.na(num.total)) num.total <- Inf
    stopifnot(is.null(save.gdsfn) | is.character(save.gdsfn))
    if (is.character(save.gdsfn))
        stopifnot(!anyNA(save.gdsfn), length(save.gdsfn)==1L)
    stopifnot(is.logical(verbose), length(verbose)==1L)

    # check packages
    if (!suppressPackageStartupMessages(requireNamespace("SNPRelate")))
        stop("The package 'SNPRelate' should be installed.")

    stopifnot(is.numeric(seed), length(seed)==1L, is.finite(seed))
    if (!is.null(seed)) set.seed(seed)
    if (is.null(parallel.multi)) parallel.multi <- parallel

    is_multi_fn <- is.character(gdsfile) && length(gdsfile)>1L
    if (verbose && !is_multi_fn)
        cat(.crayon_inverse("Perform linkage disequilibrium (LD) pruning:\n"))

    # GDS file(s)
    if (is_multi_fn)
    {
        if (anyDuplicated(gdsfile))
            stop("'gdsfile' has duplicated file name(s).")
        if (anyNA(gdsfile))
            stop("'gdsfile' has NA.")
        num.marker.each <- ceiling(num.marker / length(gdsfile))
        # for-each file name
        ans <- lapply(seq_along(gdsfile), function(i)
        {
            if (verbose && i>1L)
                .cat(paste(rep("=", 72L), collapse=""))
            # reset memory GC to reduce memory usage
            gc(verbose=FALSE, reset=TRUE, full=TRUE)
            # run LD pruning
            v <- seqFitLDpruning(gdsfile[i], sample.id=sample.id,
                ld.threshold=ld.threshold, maf=maf, missing.rate=missing.rate,
                autosome.only=autosome.only, use.cateMAC=use.cateMAC,
                num.marker=num.marker.each, num.total=Inf,
                parallel=parallel, verbose=verbose)
            if (is.null(names(v))) names(v) <- rep("", length(v))
            v
        })
        # check num.total
        ns <- vapply(ans, function(v) sum(names(v)==""), 0L)
        if (sum(ns) > num.total)
        {
            d <- data.frame(
                g = rep(1:length(ns), times=ns),
                i = unlist(lapply(ans, function(v) which(names(v)==""))))
            if (verbose)
            {
                .cat(paste(rep("=", 72L), collapse=""))
                .cat("Randomly select ", num.total, " from ", nrow(d),
                    " LD-pruned variants.")
            }
            d <- d[sort(sample.int(nrow(d), num.total)), ]
            ans <- lapply(seq_along(ns), function(j)
            {
                ii <- ans[[j]]
                jj <- d$i[d$g==j]
                c(ii[names(ii)!=""], ii[jj])
            })
        }
        names(ans) <- gdsfile
        # merging process
        if (!is.null(save.gdsfn))
        {
            if (verbose)
                .cat(.crayon_inverse(paste(c("\n", rep("=",72L)), collapse="")))
            # reset memory GC to reduce memory usage
            gc(verbose=FALSE, reset=TRUE, full=TRUE)
            save.gdsfn.part <- paste0(save.gdsfn, "_part", seq_along(gdsfile))
            on.exit(unlink(save.gdsfn.part, force=TRUE))

            # in parallel if possible
            seqParApply(parallel.multi, seq_along(gdsfile),
                FUN = function(i, snplst, gdsfn, sampid, verbose)
            {
                outfn <- paste0(gdsfn, "_part", i)
                if (verbose)
                {
                    if (sink.number() > 0) sink()  # TODO
                    cat(paste0("Export to ", sQuote(basename(outfn)),
                        " [start at ", .tm(), "]\n"))
                }
                f <- seqOpen(names(snplst)[i])
                on.exit(seqClose(f))
                seqSetFilter(f, sample.id=sampid, variant.id=snplst[[i]],
                    warn=FALSE, verbose=FALSE)
                seqExport(f, outfn, info.var=character(), fmt.var=character(),
                    samp.var=character(), verbose=FALSE)
                NULL  # no output
            }, snplst=ans, gdsfn=save.gdsfn, sampid=sample.id, verbose=verbose)

            # merge files
            if (verbose)
                .cat(.crayon_inverse(paste(rep("=",72L), collapse="")))
            seqMerge(save.gdsfn.part, save.gdsfn, verbose=verbose)
        }
        # output
        return(invisible(ans))

    } else if (is.character(gdsfile))
    {
        if (verbose)
            .cat("    open ", sQuote(gdsfile))
        gdsfile <- seqOpen(gdsfile, allow.duplicate=TRUE)
        on.exit(seqClose(gdsfile))
    } else {
        # save the filter on GDS file
        seqFilterPush(gdsfile)
        on.exit(seqFilterPop(gdsfile))
    }

    nthread <- 1L
    if (is.numeric(parallel)) nthread <- parallel

    # run LD pruning using SNPRelate
    snpset <- SNPRelate::snpgdsLDpruning(gdsfile, sample.id, variant.id,
        autosome.only=autosome.only, maf=maf, missing.rate=missing.rate,
        ld.threshold=ld.threshold, num.thread=nthread, verbose=verbose)
    nchrom <- length(snpset)
    snpset <- unlist(snpset, use.names=FALSE)
    if (length(snpset) > num.total)
    {
        if (verbose)
        {
            .cat("Randomly select ", num.total, " from ",
                length(snpset), " variants.")
        }
        snpset <- snpset[sort(sample.int(length(snpset), num.total))]
    }

    # need variants for MAC categories
    if (!isFALSE(use.cateMAC))
    {
        if (verbose)
            cat("Select ultra rare variants ...\n")
        .show_use_cateMAC(use.cateMAC, verbose)
        seqSetFilter(gdsfile, sample.id=sample.id, verbose=FALSE)
        seqSetFilter(gdsfile, variant.id=variant.id, verbose=FALSE)
        ii <- seqGetData(gdsfile, "$variant_index")
        v <- seqGetAF_AC_Missing(gdsfile, minor=TRUE, parallel=parallel,
            verbose=verbose)
        mac <- v$ac
        # missing rates
        if (is.finite(missing.rate))
        {
            x <- v$miss <= missing.rate
            x[is.na(x)] <- FALSE
            ii <- ii[x]; mac <- mac[x]
            if (verbose)
            {
                .cat("    excluding ", .pretty(sum(!x)),
                    " variants according missing rates")
            }
        }
        remove(v)  # remove unused v
        # MAC categories
        var_ii <- NULL; nm <- NULL
        low <- 0
        for (u in use.cateMAC)
        {
            s <- paste0("[", low, ",", u, ")")
            x <- mac < u
            x[is.na(x)] <- FALSE
            if (any(x))
            {
                iix <- ii[x]
                n <- min(num.marker, length(iix))
                var_ii <- c(var_ii, sort(sample(iix, n)))
                nm <- c(nm, rep(s, n))
                ii <- ii[!x]; mac <- mac[!x]
            }
            if (verbose)
            {
                .cat("    ", s, "\trandom ", num.marker, " from ", sum(x),
                    " variants")
            }
            low <- u
        }
        # finally
        seqSetFilter(gdsfile, variant.sel=var_ii, warn=FALSE, verbose=FALSE)
        snpset <- unique(c(seqGetData(gdsfile, "variant.id"), snpset))
        names(snpset) <- c(nm, rep("", length(snpset)-length(nm)))
    }

    # output to a GDS file?
    if (is.character(save.gdsfn))
    {
        seqSetFilter(gdsfile, variant.id=snpset, warn=FALSE, verbose=FALSE)
        seqExport(gdsfile, save.gdsfn, info.var=character(),
            fmt.var=character(), samp.var=character(), verbose=verbose)
    } else {
        if (verbose && nchrom>1L)
            .cat("Done (", length(snpset), " variants in total).")
    }

    invisible(snpset)
}



#######################################################################
# Calculation of GRM

.simd_calc_cpuinfo <- function() .Call(saige_simd_sp_grm)

.fit_calc_sp_grm <- function(gdsfile, nsnp.sub.random, maf, missing.rate,
    rel.cutoff, num.thread, return.ID, verbose, verbose_progress)
{
    dm <- seqSummary(gdsfile, "genotype", verbose=FALSE)$seldim
    nsamp <- dm[2L]
    nvar <- nvar_tot <- dm[3L]
    if ((nsnp.sub.random > 0L) && (nvar > nsnp.sub.random) && (nvar > 1L))
    {
        # random indices
        ii <- sort(sample.int(nvar, nsnp.sub.random))
        nvar <- nsnp.sub.random
    }
    if (verbose)
    {
        cat("Genotype matrix:\n")
        .cat("    # of samples: ", .pretty(nsamp))
        if (nvar == nvar_tot)
        {
            .cat("    # of variants: ", .pretty(nvar))
        } else {
            .cat("    # of variants: ", .pretty(nvar),
                " randomly selected from the full set (",
                .pretty(nvar_tot), ")")
        }
        .cat("    MAF threshold: >= ", maf)
        .cat("    missing genotype threshold: <= ", missing.rate)
        .cat("    # of threads: ", num.thread)
        .cat("    relatedness threshold: ", rel.cutoff)
        .cat("    using CPU capability: ", .simd_calc_cpuinfo())
    }
    if (nsamp <= 0L || nvar <= 0L)
        stop("No selected sample or variant.")

    # fill 2-bit genotypes
    if (verbose_progress)
    {
        if (nvar < nvar_tot)
        {
            cat(">>>>  First step (sparse GRM)  <<<<\n")
            cat("Loading the genotypes (the full SNP set):\n")
        } else
            cat("Loading the genotypes:\n")
    }
    nr <- ceiling(nvar_tot / 4L)  # in bytes
    ext_nb <- ceiling(nr/4L)*4L - nr  # 32-bit aligned
    g_pack <- seqGet2bGeno(gdsfile, samp_by_var=FALSE, ext_nbyte=ext_nb,
        verbose=verbose_progress)
    g_pack2 <- g_pack
    if (verbose_progress)
    {
        cat("    ", nrow(g_pack), " x ", ncol(g_pack), ": ", sep="")
        print(object.size(g_pack))
    }
    # randomly selected SNPs
    if (nvar < nvar_tot)
    {
        # rearrange, the randomly selected SNPs are stored before non-selection
        j <- rep(1L, nrow(g_pack)*4L)
        j[ii] <- 0L
        j <- order(j) - 1L  # ZERO-based
        buf <- raw(nrow(g_pack))
        .Call(saige_grm_sp_reraw, g_pack, j, buf)
        remove(buf)
        # get a new RAW matrix
        nr <- ceiling(nvar/4L)   # in bytes
        nr <- ceiling(nr/4L)*4L  # 32-bit aligned
        if (nr < nrow(g_pack))
            g_pack2 <- g_pack[seq_len(nr), ]
    }

    # initialize
    g_lookup <- matrix(NaN, nrow=8L, ncol=nrow(g_pack)*4L)
    bl_size <- 256L
    n <- ceiling(nsamp / bl_size)
    n <- n*(n+1L)/2L    # total number of blocks

    # calculation
    if (verbose_progress)
    {
        if (nvar < nvar_tot)
            .cat("Calculating GRM from the reduced SNP set (m=", nvar, "):")
        else
            cat("Calculating GRM:\n")
    }
    prog_func <- SeqArray:::.seqProgForward
    prog <- if (verbose_progress) SeqArray:::.seqProgress(n, 1L) else NULL
    v <- .Call(saige_grm_sp_calc, nvar, g_pack2, g_lookup, rel.cutoff,
        bl_size, prog, prog_func)
    remove(prog)

    # using the full set for non-zero entries
    n <- length(v$i)
    if (verbose_progress)
    {
        .cat("# of non-zero entries in GRM: ", .pretty(n))
        cat(">>>>  Second step (sparse GRM)  <<<<\n")
        .cat("Calculating the non-zero entries using the full SNP set (m=",
            nvar_tot, "):")
    }
    bl_size <- 1024L
    n <- ceiling(n/bl_size)
    prog <- if (verbose_progress) SeqArray:::.seqProgress(n, 1L) else NULL
    # v$x will be updated
    v$x <- .Call(saige_grm_sp_calc_ijx, v$i, v$j, nvar_tot,
        g_pack, g_lookup, bl_size, prog, prog_func)
    remove(prog)

    # output
    b <- any(v$x < rel.cutoff)
    if (is.na(b)) b <- FALSE
    if (b)
    {
        z <- v$x >= rel.cutoff
        m <- sparseMatrix(i=v$i[z], j=v$j[z], x=v$x[z], dims=c(nsamp, nsamp),
            symmetric=TRUE, index1=FALSE)
    } else {
        m <- sparseMatrix(i=v$i, j=v$j, x=v$x, dims=c(nsamp, nsamp),
            symmetric=TRUE, index1=FALSE)
    }

    samp_id <- seqGetData(gdsfile, "sample.id")
    colnames(m) <- rownames(m) <- samp_id
    if (isTRUE(return.ID))
    {
        m <- list(sample.id = samp_id,
            variant.id = seqGetData(gdsfile, "variant.id"))
        if (nvar < nvar_tot)
            m$variant.sub.id <- m$variant.id[ii]
        m$grm <- m
    }
    m
}

# create a sparse genetic relationship matrix
seqFitSparseGRM <- function(gdsfile, sample.id=NULL, variant.id=NULL,
    nsnp.sub.random=2000L, rel.cutoff=0.125, maf=0.01, missing.rate=0.005,
    num.thread=1L, return.ID=FALSE, seed=200L, verbose=TRUE)
{
    stopifnot(inherits(gdsfile, "SeqVarGDSClass") | is.character(gdsfile))
    stopifnot(is.numeric(nsnp.sub.random), length(nsnp.sub.random)==1L,
        nsnp.sub.random>=0L)
    stopifnot(is.numeric(rel.cutoff), length(rel.cutoff)==1L)
    if (is.na(rel.cutoff)) rel.cutoff <- -Inf
    stopifnot(is.numeric(maf), length(maf) %in% c(1L,2L))
    stopifnot(is.numeric(missing.rate), length(missing.rate)==1L)
    stopifnot(is.numeric(num.thread), length(num.thread)==1L)
    stopifnot(is.logical(return.ID), length(return.ID)==1L)
    stopifnot(is.logical(verbose), length(verbose)==1L)

    if (verbose)
        .cat(.crayon_inverse("Genetic Relationship Matrix (GRM):"))
    if (nsnp.sub.random > 0L)
    {
        nsnp.sub.random <- as.integer(floor(nsnp.sub.random/4) * 4L)
        if (nsnp.sub.random < 1L) nsnp.sub.random <- 4L
    }
    if (!is.null(seed)) set.seed(seed)

    # GDS file
    if (is.character(gdsfile))
    {
        if (verbose)
            .cat("    open ", sQuote(gdsfile))
        gdsfile <- seqOpen(gdsfile, allow.duplicate=TRUE)
        on.exit(seqClose(gdsfile))
    } else {
        # save the filter on GDS file
        seqFilterPush(gdsfile)
        on.exit(seqFilterPop(gdsfile))
    }

    # set the number of internal threads
    if (is.na(num.thread) || num.thread < 1L)
        num.thread <- 1L
    .Call(saige_set_numthread, num.thread)

    # set sample & variant filters
    if (!is.null(sample.id))
        seqSetFilter(gdsfile, sample.id=sample.id, verbose=FALSE)
    if (!is.null(variant.id))
        seqSetFilter(gdsfile, variant.id=variant.id, verbose=FALSE)
    if (verbose)
        cat("Filtering variants:\n")
    seqSetFilterCond(gdsfile, maf=maf, missing.rate=missing.rate,
        parallel=num.thread, .progress=verbose, verbose=FALSE)

    # calculating ...
    m <- .fit_calc_sp_grm(gdsfile, nsnp.sub.random, maf, missing.rate,
        rel.cutoff, num.thread, return.ID, verbose, verbose)
    # output
    if (verbose) .cat(.crayon_inverse("Done."))
    m
}


# create a dense genetic relationship matrix
seqFitDenseGRM <- function(gdsfile, sample.id=NULL, variant.id=NULL,
    maf=0.01, missing.rate=0.005, num.thread=1L, use.double=TRUE,
    return.ID=FALSE, verbose=TRUE)
{
    stopifnot(inherits(gdsfile, "SeqVarGDSClass") | is.character(gdsfile))
    stopifnot(is.numeric(maf), length(maf) %in% c(1L,2L))
    stopifnot(is.numeric(missing.rate), length(missing.rate)==1L)
    stopifnot(is.numeric(num.thread), length(num.thread)==1L)
    stopifnot(is.logical(return.ID), length(return.ID)==1L)
    stopifnot(is.logical(use.double), length(use.double)==1L)
    stopifnot(is.logical(verbose), length(verbose)==1L)
    if (verbose)
        .cat(.crayon_inverse("Genetic Relationship Matrix (GRM):"))

    # GDS file
    if (is.character(gdsfile))
    {
        if (verbose)
            .cat("    open ", sQuote(gdsfile))
        gdsfile <- seqOpen(gdsfile, allow.duplicate=TRUE)
        on.exit(seqClose(gdsfile))
    } else {
        # save the filter on GDS file
        seqFilterPush(gdsfile)
        on.exit(seqFilterPop(gdsfile))
    }

    # set the number of internal threads
    if (is.na(num.thread) || num.thread < 1L)
        num.thread <- 1L
    .Call(saige_set_numthread, num.thread)

    # set sample & variant filters
    if (!is.null(sample.id))
        seqSetFilter(gdsfile, sample.id=sample.id, verbose=FALSE)
    if (!is.null(variant.id))
        seqSetFilter(gdsfile, variant.id=variant.id, verbose=FALSE)
    if (verbose)
        cat("Filtering variants:\n")
    seqSetFilterCond(gdsfile, maf=maf, missing.rate=missing.rate,
        parallel=num.thread, .progress=verbose, verbose=FALSE)
    dm <- seqSummary(gdsfile, "genotype", verbose=FALSE)$seldim
    nsamp <- dm[2L]
    nvar <- dm[3L]
    if (verbose)
    {
        cat("Genotype matrix:\n")
        .cat("    # of samples: ", .pretty(nsamp))
        .cat("    # of variants: ", .pretty(nvar))
        .cat("    MAF threshold: >= ", maf)
        .cat("    missing genotype threshold: <= ", missing.rate)
        .cat("    # of threads: ", num.thread)
        .cat("    using CPU capability: ", .simd_calc_cpuinfo())
    }
    if (nsamp <= 0L || nvar <= 0L)
        stop("No selected sample or variant.")

    # fill 2-bit genotypes
    if (verbose)
        cat("Loading the genotypes:\n")
    nr <- ceiling(nvar / 4L)  # in bytes
    ext_nb <- ceiling(nr/4L)*4L - nr  # 32-bit aligned
    g_pack <- seqGet2bGeno(gdsfile, samp_by_var=FALSE, ext_nbyte=ext_nb,
        verbose=verbose)
    if (verbose)
    {
        cat("    ", nrow(g_pack), " x ", ncol(g_pack), ": ", sep="")
        print(object.size(g_pack))
    }

    # initialize
    g_lookup <- matrix(NaN, nrow=8L, ncol=nrow(g_pack)*4L)
    bl_size <- 256L
    n <- ceiling(nsamp / bl_size)
    n <- n*(n+1L)/2L    # total number of blocks

    # calculation
    if (verbose)
        cat("Calculating dense GRM:\n")
    prog_func <- SeqArray:::.seqProgForward
    prog <- if (verbose) SeqArray:::.seqProgress(n, 1L) else NULL
    m <- .Call(saige_grm_ds_calc, nvar, g_pack, g_lookup, use.double, bl_size,
        prog, prog_func)
    remove(prog)

    # output
    samp_id <- seqGetData(gdsfile, "sample.id")
    colnames(m) <- rownames(m) <- samp_id
    if (isTRUE(return.ID))
    {
        m <- list(sample.id = samp_id,
            variant.id = seqGetData(gdsfile, "variant.id"),
            grm = m)
    }
    if (verbose) .cat(.crayon_inverse("Done."))
    m
}

#######################################################################
#
# Package name: SAIGEgds
#
# Description:
#     Scalable and accurate implementation of generalized mixed models
# using GDS files
#
# Copyright (C) 2019-2024    Xiuwen Zheng / AbbVie-ComputationalGenomics
# License: GPL-3
#


#######################################################################
# Beta parameters in aggregate tests
#

AggrParamBeta <- structure(c(1,1,1,25), dim=c(2L,2L),
    dimnames=list(c("shape1", "shape2"), c("b_1_1", "b_1_25")))

.check_wbeta <- function(wbeta)
{
    stopifnot(is.numeric(wbeta), is.finite(wbeta))
    err <- "'wbeta' should be a length-two vector or a matrix with two rows."
    if (is.vector(wbeta))
    {
        if (length(wbeta) != 2L) stop(err)
    } else if (is.matrix(wbeta))
    {
        if (NCOL(wbeta) <= 0L) stop(err)
    }
    invisible()
}

.show_wbeta <- function(wbeta, verbose)
{
    if (verbose)
    {
        v <- apply(wbeta, 2L, function(x)
                paste0("beta(", x[1L], ",", x[2L], ")"))
        .cat("    variant weights: ", paste(v, collapse=", "))
    }
    invisible()
}

.set_check_sample_id <- function(gdsfile, modobj)
{
    seqSetFilter(gdsfile, sample.id=modobj$sample.id, warn=FALSE, verbose=FALSE)
    sid <- seqGetData(gdsfile, "sample.id")
    if (length(sid) != length(modobj$sample.id))
        stop("Some of sample IDs are not available in the GDS file.")
    ii <- match(sid, modobj$sample.id)
    if (any(is.na(ii)))
        stop("Sample IDs do not match.")
    ii
}

.set_show_units <- function(gdsfile, mobj, units, spa.pval, var.ratio, verbose)
{
    seqSetFilter(gdsfile, variant.sel=unlist(units$index), warn=FALSE,
        verbose=FALSE)
    dm <- seqSummary(gdsfile, "genotype", verbose=FALSE)$seldim
    if (dm[2L] <= 0)
        stop("No sample in the genotypic data set!")
    if (dm[3L] <= 0)
        stop("No variant in the genotypic data set!")
    v <- lengths(units$index, use.names=FALSE)
    sz_wmax <- max(v)
    if (verbose)
    {
        .cat("    # of samples: ", .pretty(dm[2L]))
        .cat("    # of variants in total: ", .pretty(dm[3L]))
        .cat("    trait: ", mobj$trait.type)
        .cat("Input sets of variants:")
        .cat("    # of units: ", .pretty(length(units$index)))
        .cat("    avg # of variants per unit: ", mean(v))
        .cat("    min # of variants in a unit: ", min(v))
        .cat("    max # of variants in a unit: ", sz_wmax)
        .cat("    sd  # of variants in a unit: ", sd(v))
        .cat("Parameters:")
        if (length(var.ratio) == 1L)
        {
            .cat("    variance ratio for approximation: ", var.ratio)
        } else {
            cat("    variance ratio for approximation (MAC categories):\n")
            v <- var.ratio
            attr(v, "cateMAC") <- NULL
            print(v, width=1024L)
        }
    }
    sz_wmax
}

.show_maf <- function(gdsfile, parallel)
{
    cat("Calculating minor allele frequencies (MAF):\n")
    maf <- seqAlleleFreq(gdsfile, minor=TRUE, verbose=TRUE, parallel=parallel)
    cat(sprintf(
        "    MAF: avg (%.5f), min (%.5f), max (%.5f), sd (%.5f)\n",
        mean(maf, na.rm=TRUE), min(maf, na.rm=TRUE),
        max(maf, na.rm=TRUE), sd(maf, na.rm=TRUE)))
    invisible()
}

.mapply <- function(mat_lst, irow)
{
    unlist(lapply(mat_lst, function(m) m[irow, ]))
}

.ncol <- function(x) if (length(d <- dim(x)) > 1L) d[2L] else 0L

.aggr_ret_obj <- function(units, obj, wbeta)
{
    # data.frame
    ans <- units$desp
    if (!is.null(ans$numvar)) ans$numvar <- NULL  # remove numvar
    ns <- vapply(obj, .ncol, 0L)
    ans <- lapply(ans, function(v) rep(v, times=ns))
    ans <- as.data.frame(ans)
    # set-based variables
    ans$maxMAF <- .mapply(obj, 1L)
    ans$numvar <- as.integer(.mapply(obj, 2L))
    ans$macmin <- .mapply(obj, 3L)
    ans$macmed <- .mapply(obj, 4L)
    ans$macmax <- .mapply(obj, 5L)
    ans$summac <- .mapply(obj, 6L)
    # weight beta
    v <- as.integer(.mapply(obj, 7L))
    attr(v, "levels") <- c(sprintf("(%g,%g)", wbeta[1L,], wbeta[2L,]), "Cauchy")
    attr(v, "class") <- "factor"
    ans$weight <- v
    # output
    ans
}

.aggr_ret_gds <- function(outf, gdsfile, units, obj, wbeta, Add)
{
    # add attributes for file format
    put.attr.gdsn(outf$root, "FileFormat", "SAIGE_OUTPUT_SET")
    put.attr.gdsn(outf$root, "Version",
        paste0("SAIGEgds_", packageVersion("SAIGEgds")))
    # add sample IDs
    Add("sample.id", seqGetData(gdsfile, "sample.id"))
    # data.frame
    ans <- units$desp
    if (!is.null(ans$numvar)) ans$numvar <- NULL  # remove numvar
    ns <- vapply(obj, .ncol, 0L)
    ans <- lapply(ans, function(v) rep(v, times=ns))
    for (nm in names(ans)) Add(nm, ans[[nm]])
    # set-based variables
    Add("maxMAF", .mapply(obj, 1L))
    Add("numvar", as.integer(.mapply(obj, 2L)))
    Add("macmin", .mapply(obj, 3L))
    Add("macmed", .mapply(obj, 4L))
    Add("macmax", .mapply(obj, 5L))
    Add("summac", .mapply(obj, 6L))
    # weight beta
    v <- as.integer(.mapply(obj, 7L))
    attr(v, "levels") <- c(sprintf("(%g,%g)", wbeta[1L,], wbeta[2L,]), "Cauchy")
    attr(v, "class") <- "factor"
    Add("weight", v)
    invisible()
}


#######################################################################
# SAIGE burden tests
#

seqAssocGLMM_Burden <- function(gdsfile, modobj, units, maxMAF=0.01,
    wbeta=AggrParamBeta, missing=0.05, ccimb.adj=TRUE, ER.mac=4.5, dsnode="",
    res.savefn="", res.compress="ZIP", parallel=FALSE,
    verbose=TRUE, verbose.maf=FALSE)
{
    stopifnot(inherits(gdsfile, "SeqVarGDSClass") | is.character(gdsfile))
    stopifnot(inherits(units, "SeqUnitListClass"))
    stopifnot(is.numeric(maxMAF), 0<maxMAF & maxMAF<=1)
    .check_wbeta(wbeta)
    stopifnot(is.numeric(missing), length(missing)==1L)
    stopifnot(is.logical(ccimb.adj), length(ccimb.adj)==1L)
    stopifnot(is.numeric(ER.mac), length(ER.mac)==1L)
    stopifnot(is.character(dsnode), length(dsnode)==1L, !is.na(dsnode))
    stopifnot(is.character(res.savefn), length(res.savefn)==1L)
    .check_compress(res.compress)
    stopifnot(is.logical(verbose), length(verbose)==1L)
    stopifnot(is.logical(verbose.maf), length(verbose.maf)==1L)
    if (verbose)
        .cat(.crayon_inverse("SAIGE burden analysis:"))

    # check model
    modobj <- .check_modobj(modobj, verbose)
    var.ratio <- .get_var_ratio(modobj)
    if (!length(maxMAF)) maxMAF <- 1
    maxMAF <- sort(maxMAF, decreasing=TRUE)
    spa.pval <- NaN  # according to Cutoff=2 in SPAtest
    if (!isTRUE(ccimb.adj)) spa.pval <- -1
    if (is.na(ER.mac)) ER.mac <- 0

    # GDS file
    if (is.character(gdsfile))
    {
        if (verbose)
            .cat("    open ", sQuote(gdsfile))
        gdsfile <- seqOpen(gdsfile, allow.duplicate=TRUE)
        on.exit(seqClose(gdsfile))
    } else {
        # save the filter on GDS file
        seqFilterPush(gdsfile)
        on.exit(seqFilterPop(gdsfile))
    }

    # determine the GDS node for dosages
    dsnode <- .dsnode(gdsfile, dsnode, TRUE)

    # check sample ID
    ii <- .set_check_sample_id(gdsfile, modobj)

    # set variant filter and show summary
    sz_wmax <- .set_show_units(gdsfile, modobj, units, spa.pval, var.ratio,
        verbose)

    # show beta weights
    if (!is.matrix(wbeta))
        wbeta <- matrix(wbeta, nrow=2L)
    .show_wbeta(wbeta, verbose)
    if (verbose)
    {
        .cat("    MAF threshold", ifelse(length(maxMAF)>1L, "s", ""), ": ",
            paste(maxMAF, collapse=", "))
        .cat("    missing proportion threshold: ", .pretty_lt_eq(missing))
    }

    # update parallel object
    njobs <- SeqArray:::.NumParallel(parallel)
    parallel <- SeqArray:::.McoreParallel(parallel)
    is_fork <- SeqArray:::.IsForking(parallel)  # is forking or not?
    if (verbose)
        .cat("    # of processes: ", njobs)

    # show allele frequencies
    if (verbose && isTRUE(verbose.maf))
        .show_maf(gdsfile, parallel)

    # initialize the internal model parameters
    mobj <- .init_nullmod(modobj, ii, 0, 0, missing, spa.pval, ER.mac,
        var.ratio, 2L, modobj$Sigma_inv, modobj$chol_inv_X_Sigma, maxMAF,
        wbeta, sz_wmax)

    # initialize internally
    if (njobs<=1L || is_fork)
    {
        # forking, no need to distribute model parameters
        .Call(saige_score_test_init, mobj)
    } else {
        # pass the model parameters to each process
        if (verbose)
            cat("Distribute the model parameters to the", njobs, "processes\n")
        # initialize
        seqParallel(parallel, NULL, split="none", .combine="none",
            FUN = function(mobj) {
                eval(.load_lib)
                .PkgEnv$mobj <- mobj
                .Call(saige_score_test_init, mobj)
            }, mobj=mobj)
        # finalize
        on.exit({
            seqParallel(parallel, NULL, split="none", .combine="none",
                FUN = function() { .PkgEnv$mobj <- NULL })
        }, add=TRUE)
    }

    # scan all variant units
    if (verbose)
        cat("Calculating p-values:\n")
    rv <- seqUnitApply(gdsfile, units, dsnode,
        FUN=function(x, maxMAF) .Call(saige_burden_test_pval, x, maxMAF),
        as.is="list", parallel=parallel, .useraw=NA, .progress=verbose,
        maxMAF=maxMAF)

    # check
    if (length(rv) != length(units$index))
        stop("seqUnitApply() returns a vector of wrong length.")

    # output to a GDS file?
    isfn <- !is.na(res.savefn) && res.savefn!=""
    if (isfn && grepl("\\.gds$", res.savefn, ignore.case=TRUE))
    {
        if (verbose)
            .cat("Save to ", sQuote(res.savefn), " ...")
        # add function
        Add <- function(varnm, val)
            add.gdsn(outf, varnm, val, compress=res.compress[1L], closezip=TRUE)
        # create a GDS file
        outf <- createfn.gds(res.savefn)
        on.exit(closefn.gds(outf), add=TRUE)
        .aggr_ret_gds(outf, gdsfile, units, rv, wbeta, Add)
        Add("beta", .mapply(rv, 8L))
        Add("SE", .mapply(rv, 9L))
        Add("pval", .mapply(rv, 10L))
        Add("method", .pval_method(.mapply(rv, 11L)))
        if (modobj$trait.type == "binary")
        {
            Add("p.norm", .mapply(rv, 12L))
            Add("converged", .mapply(rv, 13L)==1L)
        }
        if (verbose) cat(.crayon_inverse("Done.\n"))
        # output nothing
        invisible()

    } else {
        # output
        ans <- .aggr_ret_obj(units, rv, wbeta)
        ans$beta <- .mapply(rv, 8L)
        ans$SE <- .mapply(rv, 9L)
        ans$pval <- .mapply(rv, 10L)
        ans$method <- .pval_method(.mapply(rv, 11L))
        if (modobj$trait.type == "binary")
        {
            ans$p.norm <- .mapply(rv, 12L)
            ans$converged <- .mapply(rv, 13L)==1L
        }
        # save file?
        .save_R_obj(ans, res.compress, res.savefn, verbose)
    }
}



#######################################################################
# SAIGE SKAT tests
#

# initialize loading SKAT functions
.skat_init_func <- function()
{
    .Call("saige_init_skat_pkg", PACKAGE="SAIGEgds")
}

# get p-value from mixed Chi square distribution
#   (try CompQuadForm first, then saddle point method, suggested by UW DCC)
# used in saige_main.cpp
.skat_eig_chiq <- function(Q, eigval)
{
    # try integration method
    v <- suppressWarnings(CompQuadForm::davies(Q, eigval, acc=1e-9))
    if ((v$ifault > 0L) || (v$Qq < 1e3*.Machine$double.eps) || (v$Qq > 1))
    {
        # try saddlepoint method
        survey:::saddle(Q, eigval)
    } else {
        v$Qq
    }
}

# SKAT tests
seqAssocGLMM_SKAT <- function(gdsfile, modobj, units, maxMAF=0.01,
    wbeta=AggrParamBeta, missing=0.05, collapse.mac=10,
    collapse.method=c("max", "sum"), ccimb.adj=TRUE, ER.mac=4.5, dsnode="",
    res.savefn="", res.compress="ZIP", parallel=FALSE,
    verbose=TRUE, verbose.maf=FALSE)
{
    stopifnot(inherits(gdsfile, "SeqVarGDSClass") | is.character(gdsfile))
    stopifnot(inherits(units, "SeqUnitListClass"))
    stopifnot(is.numeric(maxMAF), 0<maxMAF & maxMAF<=1)
    .check_wbeta(wbeta)
    stopifnot(is.numeric(missing), length(missing)==1L)
    stopifnot(is.numeric(collapse.mac), length(collapse.mac)==1L,
        is.finite(collapse.mac))
    collapse.method <- match.arg(collapse.method)
    stopifnot(is.logical(ccimb.adj), length(ccimb.adj)==1L)
    stopifnot(is.numeric(ER.mac), length(ER.mac)==1L)
    stopifnot(is.character(dsnode), length(dsnode)==1L, !is.na(dsnode))
    stopifnot(is.character(res.savefn), length(res.savefn)==1L)
    .check_compress(res.compress)
    stopifnot(is.logical(verbose), length(verbose)==1L)
    stopifnot(is.logical(verbose.maf), length(verbose.maf)==1L)

    # check packages
    pkg_cqf <- suppressPackageStartupMessages(
        requireNamespace("CompQuadForm", quietly=TRUE))
    pkg_svy <- suppressPackageStartupMessages(
        requireNamespace("survey", quietly=TRUE))
    if (!pkg_cqf || !pkg_svy)
        stop("The packages 'CompQuadForm' and 'survey' should be installed.")

    if (verbose)
        .cat(.crayon_inverse("SAIGE SKAT analysis:"))

    # check model
    modobj <- .check_modobj(modobj, verbose)
    if (is.null(modobj$Sigma_inv) || is.null(modobj$chol_inv_X_Sigma))
    {
        stop("A (sparse) genetic relationship matrix 'grm.mat' should be ",
            "specified in seqFitNullGLMM_SPA(), when the null model is built ",
            "for SKAT.")
    }

    # variance ratio
    var.ratio <- .get_var_ratio(modobj)
    spa.pval <- NaN  # according to Cutoff=2 in SPAtest
    if (!isTRUE(ccimb.adj)) spa.pval <- -1
    if (is.na(ER.mac)) ER.mac <- 0

    # GDS file
    if (is.character(gdsfile))
    {
        if (verbose)
            .cat("    open ", sQuote(gdsfile))
        gdsfile <- seqOpen(gdsfile, allow.duplicate=TRUE)
        on.exit(seqClose(gdsfile))
    } else {
        # save the filter on GDS file
        seqSetFilter(gdsfile, action="push", verbose=FALSE)
        on.exit(seqSetFilter(gdsfile, action="pop", verbose=FALSE))
    }

    # determine the GDS node for dosages
    dsnode <- .dsnode(gdsfile, dsnode, TRUE)

    # check sample ID
    ii <- .set_check_sample_id(gdsfile, modobj)

    # set variant filter and show summary
    sz_wmax <- .set_show_units(gdsfile, modobj, units, spa.pval, var.ratio,
        verbose)
    if (verbose)
    {
        .cat("    MAC threshold for collapsing ultra rare variants: <= ",
            sprintf("%.15g", collapse.mac))
        if (is.finite(collapse.mac) && collapse.mac>0)
        {
            if (collapse.method == "max")
                cat("        max: maximum of dosages of rare variants\n")
            else if (collapse.method == "sum")
                cat("        sum: sum up rare genotypes\n")
        }
        if (modobj$trait.type == "binary")
        {
            if (isTRUE(ccimb.adj))
                cat("    accounting for case-control imbalance\n")
            else
                cat("    not accounting for case-control imbalance\n")
        }
    }

    # show beta weights
    if (!is.matrix(wbeta))
        wbeta <- matrix(wbeta, nrow=2L)
    .show_wbeta(wbeta, verbose)
    if (verbose)
    {
        .cat("    MAF threshold", ifelse(length(maxMAF)>1L, "s", ""), ": ",
            paste(maxMAF, collapse=", "))
        .cat("    missing proportion threshold: ", .pretty_lt_eq(missing))
    }

    # update parallel object
    njobs <- SeqArray:::.NumParallel(parallel)
    parallel <- SeqArray:::.McoreParallel(parallel)
    is_fork <- SeqArray:::.IsForking(parallel)  # is forking or not?
    if (verbose)
        .cat("    # of processes: ", njobs)

    # get allele frequencies
    if (verbose && isTRUE(verbose.maf)) .show_maf(gdsfile, parallel)

    # initialize the internal model parameters
    mobj <- .init_nullmod(modobj, ii, 0, 0, missing, spa.pval, ER.mac,
        var.ratio, 2L, modobj$Sigma_inv, modobj$chol_inv_X_Sigma,
        maxMAF, wbeta, sz_wmax, collapse_method=collapse.method)
    # load package(s)
    .load_skat(FALSE)

    # initialize internally
    if (njobs<=1L || is_fork)
    {
        # forking, no need to distribute model parameters
        .Call(saige_score_test_init, mobj)
        # initialize SKAT
        mobj$Sigma_inv_cg <- .sp_to_dgCMatrix(mobj$Sigma_inv)
        .Call(saige_skat_test_init, mobj$Sigma_inv_cg, mobj$t_XVX_inv_XV,
            mobj$Si_X, mobj$XVX_inv_XV_X_Si_X)
        # finalize
        on.exit(.Call(saige_skat_test_done), add=TRUE)
    } else {
        # pass the model parameters to each process
        if (verbose)
            cat("Distribute the model parameters to the", njobs, "processes\n")
        # initialize
        seqParallel(parallel, NULL, split="none", .combine="none",
            FUN = function(mobj) {
                eval(.load_lib)
                mobj$Sigma_inv_cg <- .sp_to_dgCMatrix(mobj$Sigma_inv)
                .PkgEnv$mobj <- mobj
                # initialize SKAT
                .Call(saige_score_test_init, mobj)
                .Call(saige_skat_test_init, mobj$Sigma_inv_cg,
                    mobj$t_XVX_inv_XV, mobj$Si_X, mobj$XVX_inv_XV_X_Si_X)
            }, mobj=mobj)
        # finalize
        on.exit({
            seqParallel(parallel, NULL, split="none", .combine="none",
                FUN = function() {
                    .PkgEnv$mobj <- NULL
                    .Call(saige_skat_test_done)
                })
        }, add=TRUE)
    }

    # scan all variant units
    if (verbose)
        cat("Calculating p-values:\n")
    rv <- seqUnitApply(gdsfile, units, dsnode,
        FUN=function(x) .Call(saige_skat_test_pval, x), as.is="list",
        parallel=parallel, .useraw=NA, .progress=verbose)
    # check
    if (length(rv) != length(units$index))
        stop("seqUnitApply() returns a vector of wrong length.")

    # output to a GDS file?
    isfn <- !is.na(res.savefn) && res.savefn!=""
    if (isfn && grepl("\\.gds$", res.savefn, ignore.case=TRUE))
    {
        if (verbose)
            .cat("Save to ", sQuote(res.savefn), " ...")
        cm <- res.compress[1L]
        # add function
        Add <- function(varnm, val)
            add.gdsn(outf, varnm, val, compress=cm, closezip=TRUE)
        # create a GDS file
        outf <- createfn.gds(res.savefn)
        on.exit(closefn.gds(outf), add=TRUE)
        .aggr_ret_gds(outf, gdsfile, units, rv, wbeta, Add)
        Add("n_collapse", as.integer(.mapply(rv, 8L)))
        Add("g_ncol", as.integer(.mapply(rv, 9L)))
        Add("g_minMAC", .mapply(rv, 10L))
        Add("pval", .mapply(rv, 11L))
        if (verbose) cat(.crayon_inverse("Done.\n"))
        # output
        invisible()

    } else {
        # output
        ans <- .aggr_ret_obj(units, rv, wbeta)
        ans$n_collapse <- as.integer(.mapply(rv, 8L))
        ans$g_ncol <- as.integer(.mapply(rv, 9L))
        ans$g_minMAC <- .mapply(rv, 10L)
        ans$pval <- .mapply(rv, 11L)
        # save file?
        .save_R_obj(ans, res.compress, res.savefn, verbose)
    }
}



#######################################################################
# SAIGE ACAT-V tests
#

seqAssocGLMM_ACAT_V <- function(gdsfile, modobj, units, maxMAF=0.01,
    wbeta=AggrParamBeta, missing=0.05, ccimb.adj=TRUE, collapse.mac=10,
    ER.mac=4.5, dsnode="", res.savefn="", res.compress="ZIP", parallel=FALSE,
    verbose=TRUE, verbose.maf=FALSE)
{
    stopifnot(inherits(gdsfile, "SeqVarGDSClass") | is.character(gdsfile))
    stopifnot(inherits(units, "SeqUnitListClass"))
    stopifnot(is.numeric(maxMAF), 0<maxMAF & maxMAF<=1)
    .check_wbeta(wbeta)
    stopifnot(is.numeric(missing), length(missing)==1L)
    stopifnot(is.logical(ccimb.adj), length(ccimb.adj)==1L)
    stopifnot(is.numeric(collapse.mac), length(collapse.mac)==1L,
        is.finite(collapse.mac))
    stopifnot(is.numeric(ER.mac), length(ER.mac)==1L)
    stopifnot(is.character(dsnode), length(dsnode)==1L, !is.na(dsnode))
    stopifnot(is.character(res.savefn), length(res.savefn)==1L)
    .check_compress(res.compress)
    stopifnot(is.logical(verbose), length(verbose)==1L)
    stopifnot(is.logical(verbose.maf), length(verbose.maf)==1L)
    if (verbose)
        .cat(.crayon_inverse("SAIGE ACAT-V analysis:"))

    # check model
    modobj <- .check_modobj(modobj, verbose)
    var.ratio <- .get_var_ratio(modobj)
    if (!length(maxMAF)) maxMAF <- 1
    maxMAF <- sort(maxMAF, decreasing=TRUE)
    spa.pval <- NaN  # according to Cutoff=2 in SPAtest
    if (!isTRUE(ccimb.adj)) spa.pval <- -1
    if (is.na(ER.mac)) ER.mac <- 0

    # GDS file
    if (is.character(gdsfile))
    {
        if (verbose)
            .cat("    open ", sQuote(gdsfile))
        gdsfile <- seqOpen(gdsfile, allow.duplicate=TRUE)
        on.exit(seqClose(gdsfile))
    } else {
        # save the filter on GDS file
        seqFilterPush(gdsfile)
        on.exit(seqFilterPop(gdsfile))
    }

    # determine the GDS node for dosages
    dsnode <- .dsnode(gdsfile, dsnode, TRUE)

    # check sample ID
    ii <- .set_check_sample_id(gdsfile, modobj)

    # set variant filter and show summary
    sz_wmax <- .set_show_units(gdsfile, modobj, units, spa.pval, var.ratio,
        verbose)
    if (verbose)
    {
        .cat("    MAC threshold for collapsing ultra rare variants: <= ",
            sprintf("%.15g", collapse.mac))
    }

    # show beta weights
    if (!is.matrix(wbeta))
        wbeta <- matrix(wbeta, nrow=2L)
    .show_wbeta(wbeta, verbose)
    if (verbose)
    {
        .cat("    MAF threshold", ifelse(length(maxMAF)>1L, "s", ""), ": ",
            paste(maxMAF, collapse=", "))
        .cat("    missing proportion threshold: ", .pretty_lt_eq(missing))
    }

    # update parallel object
    njobs <- SeqArray:::.NumParallel(parallel)
    parallel <- SeqArray:::.McoreParallel(parallel)
    is_fork <- SeqArray:::.IsForking(parallel)  # is forking or not?
    if (verbose)
        .cat("    # of processes: ", njobs)

    # get allele frequencies
    if (verbose && isTRUE(verbose.maf)) .show_maf(gdsfile, parallel)

    # initialize the internal model parameters
    mobj <- .init_nullmod(modobj, ii, 0, 0, missing, spa.pval, ER.mac,
        var.ratio, 2L, modobj$Sigma_inv, modobj$chol_inv_X_Sigma,
        maxMAF, wbeta, sz_wmax, collapse.mac)

    # initialize internally
    if (njobs<=1L || is_fork)
    {
        # forking, no need to distribute model parameters
        .Call(saige_score_test_init, mobj)
    } else {
        # pass the model parameters to each process
        if (verbose)
            cat("Distribute the model parameters to the", njobs, "processes\n")
        # initialize
        seqParallel(parallel, NULL, split="none", .combine="none",
            FUN = function(mobj) {
                eval(.load_lib)
                .PkgEnv$mobj <- mobj
                .Call(saige_score_test_init, mobj)
            }, mobj=mobj)
        # finalize
        on.exit({
            seqParallel(parallel, NULL, split="none", .combine="none",
                FUN = function() { .PkgEnv$mobj <- NULL })
        }, add=TRUE)
    }

    # scan all variant units
    if (verbose)
        cat("Calculating p-values:\n")
    rv <- seqUnitApply(gdsfile, units, dsnode,
        FUN=function(x) .Call(saige_acatv_test_pval, x), as.is="list",
        parallel=parallel, .useraw=NA, .progress=verbose)
    # check
    if (length(rv) != length(units$index))
        stop("seqUnitApply() returns a vector of wrong length.")

    # output to a GDS file?
    isfn <- !is.na(res.savefn) && res.savefn!=""
    if (isfn && grepl("\\.gds$", res.savefn, ignore.case=TRUE))
    {
        if (verbose)
            .cat("Save to ", sQuote(res.savefn), " ...")
        cm <- res.compress[1L]
        # add function
        Add <- function(varnm, val)
            add.gdsn(outf, varnm, val, compress=cm, closezip=TRUE)
        # create a GDS file
        outf <- createfn.gds(res.savefn)
        on.exit(closefn.gds(outf), add=TRUE)
        .aggr_ret_gds(outf, gdsfile, units, rv, wbeta, Add)
        Add("n_single", as.integer(.mapply(rv, 8L)))
        Add("n_collapse", as.integer(.mapply(rv, 9L)))
        Add("pval", .mapply(rv, 10L))
        if (verbose) cat(.crayon_inverse("Done.\n"))
        # output
        invisible()

    } else {
        # output
        ans <- .aggr_ret_obj(units, rv, wbeta)
        ans$n_single <- as.integer(.mapply(rv, 8L))
        ans$n_collapse <- as.integer(.mapply(rv, 9L))
        ans$pval <- .mapply(rv, 10L)
        # save file?
        .save_R_obj(ans, res.compress, res.savefn, verbose)
    }
}



#######################################################################
# SAIGE ACAT-O tests
#

seqAssocGLMM_ACAT_O <- function(gdsfile, modobj, units, maxMAF=0.01,
    wbeta=AggrParamBeta, missing=0.05, collapse.mac=10,
    collapse.method=c("max", "sum"), ccimb.adj=TRUE, ER.mac=4.5, dsnode="",
    res.savefn="", res.compress="ZIP", parallel=FALSE,
    verbose=TRUE, verbose.maf=FALSE)
{
    stopifnot(inherits(gdsfile, "SeqVarGDSClass") | is.character(gdsfile))
    stopifnot(inherits(units, "SeqUnitListClass"))
    stopifnot(is.numeric(maxMAF), 0<maxMAF & maxMAF<=1)
    .check_wbeta(wbeta)
    stopifnot(is.numeric(missing), length(missing)==1L)
    stopifnot(is.numeric(collapse.mac), length(collapse.mac)==1L,
        is.finite(collapse.mac))
    collapse.method <- match.arg(collapse.method)
    stopifnot(is.logical(ccimb.adj), length(ccimb.adj)==1L)
    stopifnot(is.numeric(ER.mac), length(ER.mac)==1L)
    stopifnot(is.character(dsnode), length(dsnode)==1L, !is.na(dsnode))
    stopifnot(is.character(res.savefn), length(res.savefn)==1L)
    .check_compress(res.compress)
    stopifnot(is.logical(verbose), length(verbose)==1L)
    stopifnot(is.logical(verbose.maf), length(verbose.maf)==1L)

    # check packages
    pkg_cqf <- suppressPackageStartupMessages(
        requireNamespace("CompQuadForm", quietly=TRUE))
    pkg_svy <- suppressPackageStartupMessages(
        requireNamespace("survey", quietly=TRUE))
    if (!pkg_cqf || !pkg_svy)
        stop("The packages 'CompQuadForm' and 'survey' should be installed.")

    if (verbose)
        .cat(.crayon_inverse("SAIGE ACAT-O analysis:"))

    # check model
    modobj <- .check_modobj(modobj, verbose)
    if (is.null(modobj$Sigma_inv) || is.null(modobj$chol_inv_X_Sigma))
    {
        stop("A (sparse) genetic relationship matrix 'grm.mat' should be ",
            "specified in seqFitNullGLMM_SPA(), when the null model is built ",
            "for SKAT.")
    }

    # variance ratio
    var.ratio <- .get_var_ratio(modobj)
    spa.pval <- NaN  # according to Cutoff=2 in SPAtest
    if (!isTRUE(ccimb.adj)) spa.pval <- -1
    if (is.na(ER.mac)) ER.mac <- 0

    # GDS file
    if (is.character(gdsfile))
    {
        if (verbose)
            .cat("    open ", sQuote(gdsfile))
        gdsfile <- seqOpen(gdsfile, allow.duplicate=TRUE)
        on.exit(seqClose(gdsfile))
    } else {
        # save the filter on GDS file
        seqSetFilter(gdsfile, action="push", verbose=FALSE)
        on.exit(seqSetFilter(gdsfile, action="pop", verbose=FALSE))
    }

    # determine the GDS node for dosages
    dsnode <- .dsnode(gdsfile, dsnode, TRUE)

    # check sample ID
    ii <- .set_check_sample_id(gdsfile, modobj)

    # set variant filter and show summary
    sz_wmax <- .set_show_units(gdsfile, modobj, units, spa.pval, var.ratio,
        verbose)
    if (verbose)
    {
        .cat("    MAC threshold for collapsing ultra rare variants in ",
            "ACAT-V and SKAT: <= ", sprintf("%.15g", collapse.mac))
        cat("    ACAT-O p-values combine Burden, ACAT-V and SKAT\n")
        if (modobj$trait.type == "binary")
        {
            if (isTRUE(ccimb.adj))
                cat("    accounting for case-control imbalance\n")
            else
                cat("    not accounting for case-control imbalance\n")
        }
    }

    # show beta weights
    if (!is.matrix(wbeta))
        wbeta <- matrix(wbeta, nrow=2L)
    .show_wbeta(wbeta, verbose)
    if (verbose)
    {
        .cat("    MAF threshold", ifelse(length(maxMAF)>1L, "s", ""), ": ",
            paste(maxMAF, collapse=", "))
        .cat("    missing proportion threshold: ", .pretty_lt_eq(missing))
    }

    # update parallel object
    njobs <- SeqArray:::.NumParallel(parallel)
    parallel <- SeqArray:::.McoreParallel(parallel)
    is_fork <- SeqArray:::.IsForking(parallel)  # is forking or not?
    if (verbose)
        .cat("    # of processes: ", njobs)

    # get allele frequencies
    if (verbose && isTRUE(verbose.maf)) .show_maf(gdsfile, parallel)

    # initialize the internal model parameters
    mobj <- .init_nullmod(modobj, ii, 0, 0, missing, spa.pval, ER.mac,
        var.ratio, 2L, modobj$Sigma_inv, modobj$chol_inv_X_Sigma,
        maxMAF, wbeta, sz_wmax, collapse.mac)
    i <- match(collapse.method, c("max", "sum"))
    if (is.na(i)) stop("Internal error in 'collapse.method'.")
    mobj$collapse.method <- i
    # load package(s)
    .load_skat(FALSE)

    # initialize internally
    if (njobs<=1L || is_fork)
    {
        # forking, no need to distribute model parameters
        .Call(saige_score_test_init, mobj)
        # initialize SKAT
        mobj$Sigma_inv_cg <- .sp_to_dgCMatrix(mobj$Sigma_inv)
        .Call(saige_skat_test_init, mobj$Sigma_inv_cg, mobj$t_XVX_inv_XV,
            mobj$Si_X, mobj$XVX_inv_XV_X_Si_X)
        # finalize
        on.exit(.Call(saige_skat_test_done), add=TRUE)
    } else {
        # pass the model parameters to each process
        if (verbose)
            cat("Distribute the model parameters to the", njobs, "processes\n")
        # initialize
        seqParallel(parallel, NULL, split="none", .combine="none",
            FUN = function(mobj) {
                eval(.load_lib)
                mobj$Sigma_inv_cg <- .sp_to_dgCMatrix(mobj$Sigma_inv)
                .PkgEnv$mobj <- mobj
                .Call(saige_score_test_init, mobj)
                if (!is.null(mobj$Sigma_inv))
                {
                    # initialize SKAT
                    .Call(saige_skat_test_init, mobj$Sigma_inv_cg,
                        mobj$t_XVX_inv_XV, mobj$Si_X, mobj$XVX_inv_XV_X_Si_X)
                } else {
                    .Call(saige_skat_test_reset)
                }
            }, mobj=mobj)
        # finalize
        on.exit({
            seqParallel(parallel, NULL, split="none", .combine="none",
                FUN = function() {
                    .PkgEnv$mobj <- NULL
                    .Call(saige_skat_test_done)
                })
        }, add=TRUE)
    }

    # scan all variant units
    if (verbose)
        cat("Calculating p-values:\n")
    rv <- seqUnitApply(gdsfile, units, dsnode,
        FUN=function(x) .Call(saige_acato_test_pval, x), as.is="list",
        parallel=parallel, .useraw=NA, .progress=verbose)
    # check
    if (length(rv) != length(units$index))
        stop("seqUnitApply() returns a vector of wrong length.")

    # output to a GDS file?
    isfn <- !is.na(res.savefn) && res.savefn!=""
    if (isfn && grepl("\\.gds$", res.savefn, ignore.case=TRUE))
    {
        if (verbose)
            .cat("Save to ", sQuote(res.savefn), " ...")
        cm <- res.compress[1L]
        # add function
        Add <- function(varnm, val)
            add.gdsn(outf, varnm, val, compress=cm, closezip=TRUE)
        # create a GDS file
        outf <- createfn.gds(res.savefn)
        on.exit(closefn.gds(outf), add=TRUE)
        .aggr_ret_gds(outf, gdsfile, units, rv, wbeta, Add)
        Add("n_collapse", as.integer(.mapply(rv, 8L)))
        Add("pval", .mapply(rv, 9L))
        Add("p.burden", .mapply(rv, 10L))
        Add("p.skat", .mapply(rv, 11L))
        Add("p.acatv", .mapply(rv, 12L))
        Add("burden.beta", .mapply(rv, 13L))
        Add("burden.se", .mapply(rv, 14L))
        if (verbose) cat(.crayon_inverse("Done.\n"))
        # output
        invisible()

    } else {
        # output
        ans <- .aggr_ret_obj(units, rv, wbeta)
        ans$n_collapse <- as.integer(.mapply(rv, 8L))
        ans$pval <- .mapply(rv, 9L)
        ans$p.burden <- .mapply(rv, 10L)
        ans$p.skat <- .mapply(rv, 11L)
        ans$p.acatv <- .mapply(rv, 12L)
        ans$burden.beta <- .mapply(rv, 13L)
        ans$burden.se <- .mapply(rv, 14L)
        # save file?
        .save_R_obj(ans, res.compress, res.savefn, verbose)
    }
}

---
title: "Scalable Generalized Mixed Models in PheWAS using SAIGEgds"
author: "Xiuwen Zheng (Genomics Research Center, AbbVie, North Chicago, US)"
date: "Mar, 2020"
output:
    html_document:
        theme: spacelab
        toc: true
        number_sections: true
    pdf_document:
        toc: true
        toc_depth: 3
vignette: >
    %\VignetteIndexEntry{SAIGEgds Tutorial (single variant tests)}
    %\VignetteDepends{SAIGEgds}
    %\VignetteKeywords{GWAS, PheWAS, mixed models}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8}
---

.

.

```{r echo=FALSE}
options(width=110)
```

A phenome-wide association study (PheWAS) is known to be a powerful tool in discovery and replication of genetic association studies. The recent development in the UK Biobank resource with deep genomic and phenotyping data has provided unparalleled research opportunities. To reduce the computational complexity and cost of PheWAS, the SAIGE (scalable and accurate implementation of generalized mixed model [1]) method was proposed recently, controlling for case-control imbalance and sample structure in single variant association studies. However, it is still computationally challenging to analyze the associations of thousands of phenotypes with whole-genome variant data, especially for disease diagnoses using the ICD-10 codes of deep phenotyping.

Here we develop a new high-performance statistical package (SAIGEgds) for large-scale PheWAS using mixed models [2]. In this package, the SAIGE method is implemented with optimized C++ codes taking advantage of sparse structure of genotype dosages. SAIGEgds supports efficient genomic data structure (GDS) files [3] including both integer genotypes and numeric imputed dosages. Benchmarks using the UK Biobank White British genotype data (N=430K) with coronary heart disease and simulated cases, show that SAIGEgds is 5 to 6 times faster than the SAIGE R package in the steps of fitting null models and p-value calculations. When used in conjunction with high-performance computing (HPC) clusters and/or cloud resources, SAIGEgds provides an efficient analysis pipeline for biobank-scale PheWAS.

.

# Installation

* Bioconductor repository

```R
if (!requireNamespace("BiocManager", quietly=TRUE))
    install.packages("BiocManager")
BiocManager::install("SAIGEgds")
```
The `BiocManager::install()` approach may require that you build from source, i.e. `make` and compilers must be installed on your system -- see the [R FAQ](http://cran.r-project.org/faqs.html) for your operating system; you may also need to install dependencies manually.

.

.

# Examples

```{r message=FALSE}
library(SeqArray)
library(SAIGEgds)

# the genotype file in SeqArray GDS format (1000 Genomes Phase1, chromosome 22)
(geno_fn <- seqExampleFileName("KG_Phase1"))
```

## Preparing SNP data for genetic relationship matrix

```{r}
# open a SeqArray file in the package (1000 Genomes Phase1, chromosome 22)
gds <- seqOpen(geno_fn)
```

The LD pruning is provided by [snpgdsLDpruning()](https://rdrr.io/bioc/SNPRelate/man/snpgdsLDpruning.html) in the pacakge [SNPRelate](http://bioconductor.org/packages/SNPRelate):
```{r}
library(SNPRelate)

set.seed(1000)
snpset <- snpgdsLDpruning(gds)
str(snpset)

snpset.id <- unlist(snpset, use.names=FALSE)  # get the variant IDs of a LD-pruned set
head(snpset.id)
```

Create a genotype file for genetic relationship matrix (GRM) using the LD-pruned SNP set:
```{r}
grm_fn <- "grm_geno.gds"
seqSetFilter(gds, variant.id=snpset.id)

# export to a GDS genotype file without annotation data
seqExport(gds, grm_fn, info.var=character(), fmt.var=character(), samp.var=character())
```

If genotypes are split by chromosomes, `seqMerge()` in the SeqArray package can be used to combine the LD-pruned SNP sets.

```{r}
# close the file
seqClose(gds)
```

.

## Fitting the null model

A simulated phenotype data is used to demonstrate the model fitting:
```{r}
set.seed(1000)
sampid <- seqGetData(grm_fn, "sample.id")  # sample IDs in the genotype file

pheno <- data.frame(
    sample.id = sampid,
    y  = sample(c(0, 1), length(sampid), replace=TRUE, prob=c(0.95, 0.05)),
    x1 = rnorm(length(sampid)),
    x2 = rnorm(length(sampid)),
    stringsAsFactors = FALSE)
head(pheno)

grm_fn
```

```R
# null model fitting using GRM from grm_fn
glmm <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, grm_fn, trait.type="binary", sample.col="sample.id")
```

```{r echo=FALSE}
glmm <- readRDS(system.file("extdata", "v_glmm.rds", package="SAIGEgds"))
```


.

## P-value calculations

```{r}
# genetic variants stored in the file geno_fn
geno_fn
# calculate, using 2 processes
assoc <- seqAssocGLMM_SPA(geno_fn, glmm, mac=10, parallel=2)

head(assoc)

# filtering based on p-value
assoc[assoc$pval < 5e-4, ]
```

The output could be directly saved to an R object file or a GDS file:
```{r}
# save to 'assoc.gds'
seqAssocGLMM_SPA(geno_fn, glmm, mac=10, parallel=2, res.savefn="assoc.gds")
```

Open the output GDS file using the functions in the [gdsfmt](http://www.bioconductor.org/packages/gdsfmt) package:
```{r}
# open the GDS file
(f <- openfn.gds("assoc.gds"))
# get p-values
pval <- read.gdsn(index.gdsn(f, "pval"))
summary(pval)
closefn.gds(f)
```

Load association results using the function `seqSAIGE_LoadPval()` in SAIGEgds:
```{r}
res <- seqSAIGE_LoadPval("assoc.gds")
head(res)
```

.

## Manhattan and QQ plots for p-values

```{r fig.width=6, fig.height=3, fig.align='center'}
library(ggmanh)

g <- manhattan_plot(assoc, pval.colname="pval", chr.colname="chr",
    pos.colname="pos", x.label="Chromosome 22")
g
```

```{r fig.width=3, fig.height=3, fig.align='center'}
# QQ plot
qqunif(assoc$pval)
```
.

.


# Session Information

```{r}
sessionInfo()
```

```{r echo=FALSE}
unlink(c("grm_geno.gds", "assoc.gds"), force=TRUE)
```

.


# GPU Acceleration (Optional)

SAIGEgds supports optional OpenCL-based GPU acceleration for the GRM cross-product computation during null model fitting. This can significantly speed up the `seqFitNullGLMM_SPA()` step for large sample sizes with dense genotypes.

**Requirements:**

* An OpenCL-compatible GPU (NVIDIA, AMD, Intel, or Apple)
* OpenCL runtime library installed on the system (no compile-time dependency)

To enable GPU acceleration, set `use.gpu=TRUE`:
```R
glmm <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, gdsfile, trait.type="binary", use.gpu=TRUE)
```

If no compatible GPU or OpenCL runtime is found, the function automatically falls back to CPU computation. Note that sparse genotype storage (`geno.sparse=TRUE`, the default) uses an optimized CPU-only code path; GPU acceleration applies to dense 2-bit packed genotypes (`geno.sparse=FALSE`).

**GPU Memory Usage:**

The GPU memory required depends on the operation and dataset size. SAIGEgds checks that the total allocation does not exceed 80% of device global memory, and falls back to CPU if it does.

- **Cross-product (null model fitting):** $\lceil N/4 \rceil \times M$ (packed genotypes) + $4M \times 4$ bytes (standardized genotype lookup) + $N \times 4 \times 3$ bytes (input/output/dot vectors), where $N$ is the number of samples and $M$ is the number of variants. For example, $N$=400K samples and $M$=100K variants requires approximately 10 GB.
- **Dense GRM GEMV:** $N^2 \times 4$ bytes (GRM matrix in single-precision) + $N \times 4 \times 2$ bytes (input/output vectors). For example, $N$=5K requires approximately 100 MB.
- **GRM construction (sparse/dense):** $\lceil M/4 \rceil \times N$ (packed genotypes) + $4M \times 4$ bytes (standardized genotype lookup) + $256 \times 256 \times 4$ bytes (block output buffer). For $N$=400K and $M$=2K (the `nsnp.sub.random` subset), this is approximately 200 MB.

All GPU buffers use single-precision (32-bit) floating point. Results are converted back to double precision on the host.


# References

1. Zheng X, Davis J.Wade. SAIGEgds -- an efficient statistical tool for large-scale PheWAS with mixed models. *Bioinformatics* (2021). Mar;37(5):728-730. [DOI: 10.1093/bioinformatics/btaa731](http://dx.doi.org/10.1093/bioinformatics/btaa731).
2. Zhou W, Nielsen JB, Fritsche LG, Dey R, Gabrielsen ME, Wolford BN, LeFaive J, VandeHaar P, Gagliano SA, Gifford A, Bastarache LA, Wei WQ, Denny JC, Lin M, Hveem K, Kang HM, Abecasis GR, Willer CJ, Lee S. Efficiently controlling for case-control imbalance and sample relatedness in large-scale genetic association studies. *Nat Genet* (2018). Sep;50(9):1335-1341. [DOI: 10.1038/s41588-018-0184-y](https://doi.org/10.1038/s41588-018-0184-y)
3. Zheng X, Gogarten S, Lawrence M, Stilp A, Conomos M, Weir BS, Laurie C, Levine D. SeqArray -- A storage-efficient high-performance data format for WGS variant calls. *Bioinformatics* (2017). [DOI: 10.1093/bioinformatics/btx145](http://dx.doi.org/10.1093/bioinformatics/btx145).

.


# See also

[SeqArray](http://www.bioconductor.org/packages/SeqArray): Data Management of Large-scale Whole-genome Sequence Variant Calls

[SNPRelate](http://www.bioconductor.org/packages/SNPRelate): Parallel Computing Toolset for Relatedness and Principal Component Analysis of SNP Data

\name{seqAssocGLMM_SPA}
\alias{seqAssocGLMM_SPA}
\title{P-value calculation}
\description{
    P-value calculations using variance approximation and an adjustment of
Saddlepoint approximation in the mixed model framework.
}

\usage{
seqAssocGLMM_SPA(gdsfile, modobj, maf=NaN, mac=10, missing=0.05, spa=TRUE,
    ER.mac=4.5, dsnode="", geno.ploidy=2L, res.savefn="", res.compress="ZIP",
    parallel=FALSE, load.balancing=TRUE,
    verbose.pval=c(0, 5e-10, 5e-8, 5e-6, 5e-4, 1), verbose=TRUE)
}

\arguments{
    \item{gdsfile}{a SeqArray GDS filename, or a GDS object}
    \item{modobj}{an R object for SAIGE model parameters}
    \item{maf}{minor allele frequency threshold (checking >= maf), \code{NaN}
        for no filter}
    \item{mac}{minor allele count threshold (checking >= mac), \code{NaN}
        for no filter}
    \item{missing}{missing threshold for variants (checking <= \code{missing}),
        \code{NaN} for no filter}
    \item{spa}{\code{TRUE} for using the Saddlepoint approximation method for
        adjusting case-control imbalance}
    \item{ER.mac}{the maximum MAC for the variants with efficient resampling
        p-values}
    \item{dsnode}{"" for automatically searching the GDS nodes "genotype" and
        "annotation/format/DS", or use a user-defined GDS node in the file}
    \item{geno.ploidy}{specify the ploidy (2 by default); 0 or NA used for
        non-genotype data (e.g., CNV and \code{dsnode} should be specified
        for CNVs meanwhile)}
    \item{res.savefn}{an RData or GDS file name, "" for no saving}
    \item{res.compress}{the compression method for the output file, it should
        be one of ZIP, ZIP_RA, LZMA, LZMA_RA and none; see
        \code{\link[gdsfmt]{compression.gdsn}} for more details}
    \item{parallel}{\code{FALSE} (serial processing), \code{TRUE} (multicore
        processing), a numeric value for the number of cores, or other value;
        \code{parallel} is passed to the argument \code{cl} in
        \code{\link[SeqArray]{seqParallel}}, see
        \code{\link[SeqArray]{seqParallel}} for more details}
    \item{load.balancing}{load balancing for the calculation in parallel
        if \code{TRUE}}
    \item{verbose.pval}{\code{NULL} or a numeric vector for dividing p-values
        into the intervals and show the corresponding counts}
    \item{verbose}{if \code{TRUE}, show information}
}

\details{
    For more details of SAIGE algorithm, please refer to the SAIGE paper
[Zhou et al. 2018] (see the reference section).
}

\value{
    Return a \code{data.frame} with the following components if not saving to
a file:
    \item{id}{variant ID in the GDS file;}
    \item{chr}{chromosome;}
    \item{pos}{position;}
    \item{rs.id}{the RS IDs if it is available in the GDS file as a node
        "annotation/id";}
    \item{ref}{the reference allele;}
    \item{alt}{the alternative allele;}
    \item{AF.alt}{allele frequency for the alternative allele; the minor allele
        frequency is \code{pmin(AF.alt, 1-AF.alt)};}
    \item{mac}{minor allele count; the allele count for the alternative allele
        is \code{ifelse(AF.alt<=0.5, mac, 2*num-mac)};}
    \item{num}{the number of samples with non-missing genotypes;}
    \item{beta}{beta coefficient, log odds ratio if binary outcomes
        (alternative allele vs. reference allele);}
    \item{SE}{standard error for beta coefficient;}
    \item{pval}{final p-value (if the Saddlepoint approximation or efficient
        resampling is applied);}
    \item{method}{\code{"Normal"}: \code{pval} is calculated by assuming
        asymptotic normality; \code{"SPA"}: using the Saddlepoint approximation
        method; \code{"ER"}: using the efficient resampling}
    \item{p.norm}{p-values based on asymptotic normality (could be 0 if it
        is too small, e.g., \code{pnorm(-50) = 0} in R; used for checking only}
    \item{converged}{whether the SPA algorithm converges or not for p-values.}
}

\references{
    Zhou W, Nielsen JB, Fritsche LG, Dey R, Gabrielsen ME, Wolford BN,
LeFaive J, VandeHaar P, Gagliano SA, Gifford A, Bastarache LA, Wei WQ,
Denny JC, Lin M, Hveem K, Kang HM, Abecasis GR, Willer CJ, Lee S.
Efficiently controlling for case-control imbalance and sample relatedness in
large-scale genetic association studies. Nat Genet (2018). Sep;50(9):1335-1341.
}
\author{Xiuwen Zheng}
\seealso{
    \code{\link{seqAssocGLMM_SPA}}, \code{\link{seqSAIGE_LoadPval}}
}

\examples{
# open a GDS file
fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
gdsfile <- seqOpen(fn)

# load phenotype
phenofn <- system.file("extdata", "pheno.txt.gz", package="SAIGEgds")
pheno <- read.table(phenofn, header=TRUE, as.is=TRUE)
head(pheno)

# fit the null model
glmm <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, gdsfile, trait.type="binary")

# p-value calculation
a <- seqAssocGLMM_SPA(gdsfile, glmm, mac=10)
head(a)

# close the GDS file
seqClose(gdsfile)
}

\keyword{GDS}
\keyword{genetics}
\keyword{association}

\name{SAIGEgds-package}
\alias{SAIGEgds-package}
\alias{SAIGEgds}
\docType{package}
\title{Scalable Implementation of Generalized mixed models in Phenome-Wide
    Association Studies using GDS files}
\description{
    Scalable and accurate implementation of generalized mixed mode with the
support of Genomic Data Structure (GDS) files and highly optimized C++
implementation. It is designed for single variant tests in large-scale
phenome-wide association studies (PheWAS) with millions of variants and hundreds
of thousands of samples, e.g., UK Biobank genotype data, controlling for
case-control imbalance and sample structure in single variant association
studies.

    The implementation of SAIGEgds is based on the original SAIGE R package
(v0.29.4.4) [Zhou et al. 2018]
\url{https://github.com/weizhouUMICH/SAIGE/releases/tag/v0.29.4.4}. All of the
calculation with single-precision floating-point numbers in SAIGE are replaced
by the double-precision calculation in SAIGEgds. SAIGEgds also implements some
of the SPAtest functions in C to speed up the calculation of Saddlepoint
Approximation.
}
\details{
    \tabular{ll}{
        Package: \tab SAIGEgds \cr
        Type: \tab Package \cr
        License: \tab GPL version 3 \cr
    }
}
\references{
    Zheng X, Davis J.Wade. SAIGEgds -- an efficient statistical tool for
large-scale PheWAS with mixed models. *Bioinformatics* (2020).
DOI: 10.1093/bioinformatics/btaa731.

    Zhou W, Nielsen JB, Fritsche LG, Dey R, Gabrielsen ME, Wolford BN,
LeFaive J, VandeHaar P, Gagliano SA, Gifford A, Bastarache LA, Wei WQ, Denny JC,
Lin M, Hveem K, Kang HM, Abecasis GR, Willer CJ, Lee S. Efficiently controlling
for case-control imbalance and sample relatedness in large-scale genetic
association studies. *Nat Genet* (2018). Sep;50(9):1335-1341.

    Zhou W, Zhao Z, Nielsen JB, Fritsche LG, LeFaive J, Taliun SAG, Bi W,
Gabrielsen ME, Daly MJ, Neale BM, Hveem K, Abecasis GR, Willer CJ, et al.
Scalable generalized linear mixed model for region-based association tests
in large biobanks and cohorts. Nat Genet. 2020; 52: 634-9.

    Zheng X, Gogarten S, Lawrence M, Stilp A, Conomos M, Weir BS, Laurie C,
Levine D. SeqArray -- A storage-efficient high-performance data format for WGS
variant calls. *Bioinformatics* (2017). DOI: 10.1093/bioinformatics/btx145.
}
\author{
    Xiuwen Zheng \email{xiuwen.zheng@abbvie.com},
    Wei Zhou (the original author of the SAIGE R package,
\url{https://github.com/weizhouUMICH/SAIGE})
}

\examples{
# open the GDS file
fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
gdsfile <- seqOpen(fn)

# load phenotype
phenofn <- system.file("extdata", "pheno.txt.gz", package="SAIGEgds")
pheno <- read.table(phenofn, header=TRUE, as.is=TRUE)
head(pheno)

# fit the null model
glmm <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, gdsfile, trait.type="binary",
    geno.sparse=FALSE)

# p-value calculation
assoc <- seqAssocGLMM_SPA(gdsfile, glmm, mac=10)

head(assoc)

# close the GDS file
seqClose(gdsfile)
}

\keyword{GDS}
\keyword{genetics}
\keyword{association}

\name{seqAssocGLMM_Burden}
\alias{seqAssocGLMM_Burden}
\title{Burden tests}
\description{
    Set-based burden tests using mixed models.
}

\usage{
seqAssocGLMM_Burden(gdsfile, modobj, units, maxMAF=0.01, wbeta=AggrParamBeta,
    missing=0.05, ccimb.adj=TRUE, ER.mac=4.5, dsnode="", res.savefn="",
    res.compress="ZIP", parallel=FALSE, verbose=TRUE, verbose.maf=FALSE)
}

\arguments{
    \item{gdsfile}{a SeqArray GDS filename, or a GDS object}
    \item{modobj}{an R object for SAIGE model parameters}
    \item{units}{a list of units of selected variants, with S3 class
        \code{"SeqUnitListClass"} defined in the SeqArray package}
    \item{maxMAF}{a numeric vector for MAF thresholds, e.g.,
        \code{c(0.01, 0.001)} for rare variants}
    \item{wbeta}{weights for per-variant effect, using beta distribution
        \code{dbeta()} according to variant's MAF; a length-two vector, or a
        matrix with two rows for multiple beta parameters; by default, using
        beta(1,1) and beta(1,25) both}
    \item{missing}{missing threshold for variants (checking <= \code{missing}),
        \code{NaN} for no filter}
    \item{ccimb.adj}{whether adjusting for case-control imbalance or not}
    \item{ER.mac}{the maximum sum of MAC for the variants with efficient
        resampling p-values}
    \item{dsnode}{"" for automatically searching the GDS nodes "genotype" and
        "annotation/format/DS", or use a user-defined GDS node in the file}
    \item{res.savefn}{an RData or GDS file name, "" for no saving}
    \item{res.compress}{the compression method for the output file, it should
        be one of LZMA, LZMA_RA, ZIP, ZIP_RA and none}
    \item{parallel}{\code{FALSE} (serial processing), \code{TRUE} (multicore
        processing), a numeric value for the number of cores, or other value;
        \code{parallel} is passed to the argument \code{cl} in
        \code{\link[SeqArray]{seqParallel}}, see
        \code{\link[SeqArray]{seqParallel}} for more details}
    \item{verbose}{if \code{TRUE}, show information}
    \item{verbose.maf}{if \code{TRUE}, show summary of MAFs in units}
}

\details{
    \code{seqUnitFilterCond()} in the SeqArray package can be used to restrict
the variant sets \code{units} to a range of MAC and/or MAF.
}

\value{
    Return a \code{data.frame} with the following components if not saving to
a file:
    \item{chr}{chromosome;}
    \item{start}{the starting position;}
    \item{end}{the ending position;}
    \item{maxMAF}{maximum of MAFs;}
    \item{numvar}{the number of variants in the window; if
        \code{weight="Cauchy"}, it is the number of individual p-values in
        the Cauchy test}
    \item{macmin}{the minimum of MAC in the window;}
    \item{macmed}{the median of MAC in the window;}
    \item{macmax}{the maximum of MAC in the window;}
    \item{summac}{the sum of minor allele counts;}
    \item{weight}{the weighting strategy with the beta distribution parameters}
    \item{beta}{beta coefficient, log odds ratio if binary outcomes
        (alternative allele vs. reference allele);}
    \item{SE}{standard error for beta coefficient;}
    \item{pval}{final p-value (if the Saddlepoint approximation or efficient
        resampling is applied);}
    \item{method}{\code{"Normal"}: \code{pval} is calculated by assuming
        asymptotic normality; \code{"SPA"}: using the Saddlepoint approximation
        method; \code{"ER"}: using the efficient resampling}
    \item{p.norm}{p-values based on asymptotic normality (could be 0 if it
        is too small, e.g., \code{pnorm(-50) = 0} in R; used for checking only}
    \item{converged}{whether the SPA algorithm converges or not for p-values.}
}

\author{Xiuwen Zheng}
\seealso{
    \code{\link{seqAssocGLMM_ACAT_V}}, \code{\link{seqAssocGLMM_ACAT_O}},
    \code{\link{seqAssocGLMM_SKAT}}, \code{\link[SeqArray]{seqUnitFilterCond}}
}

\examples{
# open a GDS file
fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
gdsfile <- seqOpen(fn)

# load phenotype
phenofn <- system.file("extdata", "pheno.txt.gz", package="SAIGEgds")
pheno <- read.table(phenofn, header=TRUE, as.is=TRUE)
head(pheno)

# fit the null model
glmm <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, gdsfile, trait.type="binary")

# get a list of variant units for burden tests
units <- seqUnitSlidingWindows(gdsfile, win.size=500, win.shift=250)

a <- seqAssocGLMM_Burden(gdsfile, glmm, units, maxMAF=c(0.1,0.01))
head(a)

# close the GDS file
seqClose(gdsfile)
}

\keyword{GDS}
\keyword{genetics}
\keyword{association}

\name{pACAT}
\alias{pACAT}
\alias{pACAT2}
\title{Cauchy Combination Test}
\description{
    P-value calculation from Cauchy combination test.
}

\usage{
pACAT(p, w=NULL)
pACAT2(p, maf, wbeta=c(1,25))
}

\arguments{
    \item{p}{a numeric vector for p-values}
    \item{w}{weight for each p-value}
    \item{maf}{minor allele frequency for each p-value}
    \item{wbeta}{weights for per-variant effect, using beta distribution
        \code{dbeta()} according to variant's MAF}
}

\value{
    Return a single number for the combined p-value.
}

\references{
    Liu Y., Cheng S., Li Z., Morrison A.C., Boerwinkle E., Lin X.;
ACAT: A Fast and Powerful p Value Combination Method for Rare-Variant Analysis
in Sequencing Studies. Am J Hum Genetics 104, 410-421 (2019).
}
\seealso{
    \code{\link{seqFitNullGLMM_SPA}}, \code{\link{seqAssocGLMM_SPA}}
}

\examples{
p1 <- 10^-4
p2 <- 10^-5
p3 <- 10^-(3:20)
sapply(p3, function(p) pACAT(c(p1, p2, p)))

pACAT2(c(10^-4, 10^-6), c(0.01, 0.005))
}

\keyword{association}
\keyword{Cauchy}

\name{seqSAIGE_LoadPval}
\alias{seqSAIGE_LoadPval}
\title{Load the association results}
\description{
    Load the association results from an RData, RDS or GDS file.
}

\usage{
seqSAIGE_LoadPval(fn, varnm=NULL, index=NULL, verbose=TRUE)
}

\arguments{
    \item{fn}{RData, RDS or GDS file names, merging datasets if multiple files}
    \item{varnm}{NULL, or a character vector to include the column names; e.g.,
        \code{c("chr", "position", "rs.id", "ref", "alt", "pval")}}
    \item{index}{NULL, or a logical/numeric vector for a set of rows}
    \item{verbose}{if \code{TRUE}, show information}
}

\value{
    Return a \code{data.frame} including p-values.
}

\author{Xiuwen Zheng}
\seealso{
    \code{\link{seqFitNullGLMM_SPA}}, \code{\link{seqAssocGLMM_SPA}}
}

\examples{
(fn <- system.file("unitTests", "saige_pval.rds", package="SAIGEgds"))
pval <- seqSAIGE_LoadPval(fn)

names(pval)
#  [1] "id"         "chr"        "pos"        "rs.id"      "ref"
#  [6] "alt"        "AF.alt"     "AC.alt"     "num"        "beta"
# [11] "SE"         "pval"       "pval.noadj" "converged"

head(pval)
}

\keyword{GDS}
\keyword{genetics}
\keyword{association}

\name{seqFitLDpruning}
\alias{seqFitLDpruning}
\title{Linkage disequilibrium pruning}
\description{
    Construct LD-pruned SNP sets for genetic relationship matrix (GRM).
}

\usage{
seqFitLDpruning(gdsfile, sample.id=NULL, variant.id=NULL,
    ld.threshold=0.1, maf=0.01, missing.rate=0.005, autosome.only=TRUE,
    use.cateMAC=TRUE, num.marker=100L, num.total=100000L, save.gdsfn=NULL,
    seed=200L, parallel=FALSE, parallel.multi=NULL, verbose=TRUE)
}

\arguments{
    \item{gdsfile}{a SeqArray GDS file name, a GDS object for genotypes, or
        a character vector for a list of GDS file names when genotyeps are
        split by chromosomes}
    \item{sample.id}{NULL for all samples, or sample IDs in GRM}
    \item{variant.id}{a list of variant IDs, used to construct GRM}
    \item{ld.threshold}{the LD threshold; see \code{snpgdsLDpruning()}}
    \item{maf}{minor allele frequency threshold for genotypes in \code{gdsfile}
        (checking >= maf), if \code{variant.id=NULL}; \code{NaN} for no filter}
    \item{missing.rate}{threshold of missing rate (checking <= missing.rate),
        if \code{variant.id=NULL}; \code{NaN} for no filter}
    \item{autosome.only}{\code{TRUE} for using autosomes only}
    \item{use.cateMAC}{\code{FALSE}, to use a single global variance ratio;
        \code{TRUE} (equal to
        \code{use.cateMAC=c(1.5, 2.5, 3.5, 4.5, 5.5, 10.5, 20.5)}) for MAC
        categories (0, 1.5), [1.5, 2.5), ... [10.5, 20.5) and [20.5, Inf);
        or a numeric vector (strictly increasing) for unique cut points}
    \item{num.marker}{the number of SNPs used to calculate the variance ratio
        in each MAC category}
    \item{num.total}{the total number of LD-pruned variants excluding ultra
        rare variants; if the number of variants selected by the process of
        LD pruning is larger than \code{num.total}, the random set is used}
    \item{save.gdsfn}{if a file name is specified, construct a GDS genotype
        file to include all selected variants}
    \item{parallel}{\code{FALSE} (serial processing), \code{TRUE} (multicore
        processing), a numeric value for the number of cores, or other value;
        \code{parallel} is passed to the argument \code{cl} in
        \code{\link[SeqArray]{seqParallel}}, see
        \code{\link[SeqArray]{seqParallel}} for more details}
    \item{parallel.multi}{only applicable when there are multiple input files
        in \code{gdsfile} and \code{save.gdsfn} is used; multiple GDS files can
        be extracted to create a sub GDS file in parallel via
        \code{parallel.multi}; if \code{NULL}, use \code{parallel} instead;
        \code{parallel.multi} could have less CPU cores than \code{parallel}
        to reduce memory usage}
    \item{seed}{an integer passed to \code{set.seed()} as a seed for random
        numbers, or \code{NULL} for no action}
    \item{verbose}{if \code{TRUE}, show information}
}

\details{
    This function calls \code{\link[SNPRelate]{snpgdsLDpruning}} in the
SNPRelate package to perform linkage disequilibrium (LD) pruning. When
\code{use.cateMAC} is not \code{FALSE}, the ultra rare variants will be
selected according to the MAC categories, which could be used in the null
model fitting.
}

\value{
    Returns variant IDs or a list of variant IDs for multiple input GDS files.
}

\author{Xiuwen Zheng}
\seealso{
    \code{\link{seqFitNullGLMM_SPA}}, \code{\link{seqAssocGLMM_SPA}},
    \code{\link[SNPRelate]{snpgdsLDpruning}}
}

\examples{
library(SeqArray)
library(SAIGEgds)

# open a GDS file
gds_fn <- seqExampleFileName("KG_Phase1")

seqFitLDpruning(gds_fn, save.gdsfn="grm.gds")

# delete the temporary file
unlink("grm.gds", force=TRUE)
}

\keyword{GDS}
\keyword{genetics}
\keyword{association}

\name{seqFitNullGLMM_SPA}
\alias{seqFitNullGLMM_SPA}
\title{Fit the null model with GRM}
\description{
    Fit the null model in the mixed model framework with genetic relationship
matrix (GRM).
}

\usage{
seqFitNullGLMM_SPA(formula, data, gdsfile=NULL, grm.mat=NULL,
    trait.type=c("binary", "quantitative"), sample.col="sample.id", maf=0.01,
    missing.rate=0.01, max.num.snp=1000000L, variant.id=NULL,
    variant.id.varratio=NULL, nsnp.sub.random=2000L, rel.cutoff=0.125,
    inv.norm=c("residuals", "quant", "none"), use.cateMAC=FALSE,
    cateMAC.inc.maf=TRUE, cateMAC.simu=TRUE, use.offset=FALSE, X.transform=TRUE,
    tol=0.02, maxiter=20L, nrun=30L, tolPCG=1e-5, maxiterPCG=500L,
    num.marker=30L, tau.init=c(0,0), traceCVcutoff=0.0025, ratioCVcutoff=0.001,
    geno.sparse=TRUE, num.thread=1L, model.savefn="", seed=200L,
    fork.loading, parallel.loading=FALSE, verbose=TRUE)
}

\arguments{
    \item{formula}{an object of class \code{formula} (or one that can be
        coerced to that class), e.g., \code{y ~ x1 + x2}, see
        \code{\link[stats]{lm}}}
    \item{data}{a data frame for the formulas}
    \item{gdsfile}{a SeqArray GDS filename or a GDS object for genotypes used
        in the GRM calculation; see details}
    \item{grm.mat}{NULL, a user-defined GRM or an object of 'snpgdsGRMClass'
        returned from \code{SNPRelate::snpgdsGRM()}: a dense numeric matrix or
        a sparse symmetric matrix 'sparseMatrix' defined in \code{Matrix};
        \code{colnames} and \code{rownames} should be sample IDs;
        or \code{TRUE}, call \code{seqFitSparseGRM()} to get a sparse GRM;
        see details}
    \item{trait.type}{"binary" for binary outcomes, "quantitative" for
        continuous outcomes}
    \item{sample.col}{the column name of sample IDs corresponding to the GDS
        file}
    \item{maf}{minor allele frequency threshold for genotypes in \code{gdsfile}
        (checking >= maf), if \code{variant.id=NULL}; \code{NaN} for no filter}
    \item{missing.rate}{threshold of missing rate (checking <= missing.rate),
        if \code{variant.id=NULL}; \code{NaN} for no filter}
    \item{max.num.snp}{the maximum number of SNPs used, or -1 for no limit}
    \item{variant.id}{a list of variant IDs, considered to be used in GRM;
        or \code{NULL}, all variants in the GDS file to be the candidate
        variants}
    \item{variant.id.varratio}{a list of variant IDs, to be used in the
        variance ratio estimation; or \code{NULL} to use the candidate variants
        according to \code{variant.id}}
    \item{nsnp.sub.random}{used in \code{seqFitSparseGRM()} when
        \code{grm.mat=TRUE}: the number of SNP markers randomly selected from
        the candidate variants for the initial scan of relatedness}
    \item{rel.cutoff}{relatedness threshold for treating two individuals as
        unrelated (check >= rel.cutoff), \code{NaN} or \code{-Inf} for no
        threshold; only applicable when \code{grm.mat=TRUE}}
    \item{use.cateMAC}{\code{FALSE}, to use a single global variance ratio;
        \code{TRUE} (equal to
        \code{use.cateMAC=c(1.5, 2.5, 3.5, 4.5, 5.5, 10.5, 20.5)}) for MAC
        categories (0, 1.5), [1.5, 2.5), ... [10.5, 20.5) and [20.5, Inf);
        or a numeric vector (strictly increasing) for unique cut points}
    \item{cateMAC.inc.maf}{\code{TRUE} to include a MAC cut point according to
        MAF=\code{maf}, or a numeric vector for MAF; only applicable if
        \code{use.cateMAC} is not \code{FALSE}}
    \item{cateMAC.simu}{\code{TRUE} for using simulated independent
        genotypes under Hardy-Weinberg equilibrium if there are no enough
        SNP markers in the MAC category; only applicable if \code{use.cateMAC}
        is not \code{FALSE}}
    \item{inv.norm}{\code{"residuals"} (by default), perform the inverse normal
        transformation on the residuals after fitting the fixed effects;
        \code{"quant"}, perform the inverse normal transformation on the
        outcome variable directly (the same behavior as the SAIGE package);
        \code{"none"}, no transformation.
        If \code{inv.norm=TRUE}, perform \code{inv.norm="residuals"};
        if \code{inv.norm=FALSE}, it is as the same as \code{inv.norm="none"};
        See the reference for more discussion [Sofer et al., 2019]}
    \item{use.offset}{if \code{TRUE}, use the estimated residual offset to
        avoid calculating each fixed effect coefficient}
    \item{X.transform}{if \code{TRUE}, perform QR decomposition on the design
        matrix}
    \item{tol}{overall tolerance for model fitting}
    \item{maxiter}{the maximum number of iterations for model fitting;
        \code{maxiter=0} for using a fixed tau parameter in \code{tau.init}}
    \item{nrun}{the number of random vectors in the trace estimation}
    \item{tolPCG}{tolerance of PCG iterations}
    \item{maxiterPCG}{the maximum number of PCG iterations}
    \item{num.marker}{the number of SNPs used to calculate the variance ratio}
    \item{tau.init}{a 2-length numeric vector, the initial values for variance
        components, tau; for binary traits, the first element is always be set
        to 1. if \code{tau.init} is not specified, the second element will be
        0.5 for binary traits}
    \item{traceCVcutoff}{the threshold for coefficient of variation (CV) for
        the trace estimator, and the number of runs for trace estimation will
        be increased until the CV is below the threshold}
    \item{ratioCVcutoff}{the threshold for coefficient of variation (CV) for
        estimating the variance ratio, and the number of randomly selected
        markers will be increased until the CV is below the threshold}
    \item{geno.sparse}{if \code{TRUE} (by default), store the sparse structure
        for genotypes; otherwise, save genotypes in a 2-bit dense matrix;
        see details}
    \item{num.thread}{the number of threads, 1 by default}
    \item{model.savefn}{the filename of model output, R data file '.rda',
        '.RData', or '.rds'}
    \item{seed}{an integer passed to \code{set.seed()} as a seed for random
        numbers, or \code{NULL} for no action}
    \item{fork.loading}{deprecated, use \code{parallel.loading} instead}
    \item{parallel.loading}{load genotypes via parallel or not; multiple
        processes can reduce loading time of genotypes, but may increase the
        memory usage}
    \item{verbose}{if \code{TRUE}, show information}
}

\details{
    Utilizing the sparse structure of genotypes could significantly improve
the computational efficiency of model fitting, but it also increases the memory
usage.
    For more details of SAIGE algorithm, please refer to the SAIGE paper
[Zhou et al., 2018] (see the reference section).
}

\value{
    Returns a list with the following components:
    \item{coefficients}{the beta coefficients for fixed effects;}
    \item{tau}{a numeric vector of variance components 'Sigma_E' and 'Sigma_G';}
    \item{linear.predictors}{the linear fit on link scale;}
    \item{fitted.values}{fitted values from objects returned by modeling
        functions using \code{glm.fit};}
    \item{residuals}{residuals;}
    \item{cov}{covariance matrix of beta coefficients;}
    \item{converged}{whether the model is fitted or not;}
    \item{obj.noK}{internal use, returned object from the SPAtest package;}
    \item{var.ratio}{a data.frame with columns 'id' (variant.id), 'maf'
        (minor allele frequency), 'mac' (minor allele count),
        'var1' (the variance of score statistic),
        'var2' (a variance estimate without accounting for estimated random
        effects) and 'ratio' (var1/var2, estimated variance ratio for variance
        approximation);}
    \item{trait.type}{either "binary" or "quantitative";}
    \item{sample.id}{the sample IDs used in the model fitting;}
    \item{variant.id}{the variant IDs used in the model fitting.}
}

\references{
    Zhou W, Nielsen JB, Fritsche LG, Dey R, Gabrielsen ME, Wolford BN,
LeFaive J, VandeHaar P, Gagliano SA, Gifford A, Bastarache LA, Wei WQ,
Denny JC, Lin M, Hveem K, Kang HM, Abecasis GR, Willer CJ, Lee S.
Efficiently controlling for case-control imbalance and sample relatedness in
large-scale genetic association studies. Nat Genet (2018). Sep;50(9):1335-1341.

    Zhou W, Zhao Z, Nielsen JB, Fritsche LG, LeFaive J, Taliun SAG, Bi W,
Gabrielsen ME, Daly MJ, Neale BM, Hveem K, Abecasis GR, Willer CJ, et al.
Scalable generalized linear mixed model for region-based association tests
in large biobanks and cohorts. Nat Genet. 2020; 52: 634-9.
}
\author{Xiuwen Zheng}
\seealso{
    \code{\link{seqAssocGLMM_SPA}}
}

\examples{
# open a GDS file
fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
gdsfile <- seqOpen(fn)

# load phenotype
phenofn <- system.file("extdata", "pheno.txt.gz", package="SAIGEgds")
pheno <- read.table(phenofn, header=TRUE, as.is=TRUE)
head(pheno)

# fit the null model
glmm <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, gdsfile, trait.type="binary")
glmm

# close the GDS file
seqClose(gdsfile)
}

\keyword{GDS}
\keyword{genetics}
\keyword{association}

\name{seqFitSparseGRM}
\alias{seqFitSparseGRM}
\alias{seqFitDenseGRM}
\title{Sparse & dense genetic relationship matrix}
\description{
    Construct sparse and dense genetic relationship matrix (GRM).
}

\usage{
seqFitSparseGRM(gdsfile, sample.id=NULL, variant.id=NULL, nsnp.sub.random=2000L,
    rel.cutoff=0.125, maf=0.01, missing.rate=0.005, num.thread=1L,
    return.ID=FALSE, seed=200L, verbose=TRUE)
seqFitDenseGRM(gdsfile, sample.id=NULL, variant.id=NULL, maf=0.01,
    missing.rate=0.005, num.thread=1L, use.double=TRUE, return.ID=FALSE,
    verbose=TRUE)
}

\arguments{
    \item{gdsfile}{a SeqArray GDS file name or a GDS object for genotypes used
        in the GRM calculation; see details}
    \item{sample.id}{NULL for all samples, or sample IDs in GRM}
    \item{variant.id}{candidate variant IDs used for constructing GRM; or
        \code{NULL} for using all available candidate variants}
    \item{nsnp.sub.random}{the number of SNP markers randomly selected from the
        candidate variants for the initial scan of relatedness;
        if \code{nsnp.sub.random=0}, use all candidate SNPs}
    \item{rel.cutoff}{relatedness threshold for treating two individuals as
        unrelated (check >= rel.cutoff); \code{NaN} or \code{-Inf} for no
        threshold}
    \item{maf}{minor allele frequency threshold for genotypes in \code{gdsfile}
        (checking >= maf), if \code{variant.id=NULL}; \code{NaN} for no filter}
    \item{missing.rate}{threshold of missing rate (checking <= missing.rate),
        if \code{variant.id=NULL}; \code{NaN} for no filter}
    \item{num.thread}{the number of threads, 1 by default}
    \item{use.double}{\code{TRUE} for using 64-bit double precision to
        calculate GRM; otherwise, to use 32-bit single precision}
    \item{return.ID}{if \code{TRUE}, return the IDs for samples, the full
        variant set and the subset of variants}
    \item{seed}{an integer passed to \code{set.seed()} as a seed for random
        numbers, or \code{NULL} for no action}
    \item{verbose}{if \code{TRUE}, show information}
}

\details{
    The genetic relationship matrix (GRM) is defined as
$grm_ij = avg_l [(g_il - 2*p_l)*(g_jl - 2*p_l) / 2*p_l*(1 - p_l)]$ for
individuals i,j and locus l, where $g_il$ is 0, 1 or 2, and $p_l$ is the
allele frequency at locus l. The missing genotypes are dropped from the
calculation.
}

\value{
    If \code{return.ID=TRUE}, returns a list with \code{sample.id} for
sample IDs, \code{variant.id} for the full set of variants,
\code{variant.sub.id} for the subset of variants, and the GRM matrix.
    Otherwise, it returns a sparse or dense symmetric matrix for GRM, with
sample IDs in \code{colnames()} and \code{rownames()}.
}

\author{Xiuwen Zheng}
\seealso{
    \code{\link{seqFitNullGLMM_SPA}}, \code{\link{seqFitLDpruning}}
}

\examples{
library(Matrix)

# open a GDS file
gds_fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
gdsfile <- seqOpen(gds_fn)

seqSetFilter(gdsfile, variant.sel=1:100)
m <- seqFitSparseGRM(gdsfile, rel.cutoff=0.125)
is(m)
nnzero(m)                 # num of non-zero
nnzero(m) / prod(dim(m))  # percentage of non-zero

m <- seqFitDenseGRM(gdsfile)
str(m)

# close the GDS file
seqClose(gdsfile)
}

\keyword{GDS}
\keyword{genetics}
\keyword{association}

\name{glmmHeritability}
\alias{glmmHeritability}
\title{Heritability estimation}
\description{
    Get the heritability estimate from the SAIGE model.
}

\usage{
glmmHeritability(modobj, adjust=TRUE)
}

\arguments{
    \item{modobj}{an R object for SAIGE model parameters}
    \item{adjust}{if \code{TRUE} and binary outcomes, uses adjusted tau
        estimate for the heritability estimation}
}

\details{
    In SAIGE, penalized quasi-likelihood (PQL) is used to estimate the variance
component parameter tau. It is known to produce biased estimate of the variance
component tau using PQL. If \code{adjust=TRUE} for binary outcomes, tau is
adjusted based prevalence and observed tau using the data in Supplementary
Table 7 (Zhou et al. 2018) to reduce the bias of PQL estimate of variance
component.
}

\value{
    Return a liability scale heritability.
}

\author{Xiuwen Zheng}
\seealso{
    \code{\link{seqFitNullGLMM_SPA}}
}

\examples{
# open a GDS file
fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
gdsfile <- seqOpen(fn)

# load phenotype
phenofn <- system.file("extdata", "pheno.txt.gz", package="SAIGEgds")
pheno <- read.table(phenofn, header=TRUE, as.is=TRUE)
head(pheno)

# fit the null model
glmm <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, gdsfile, trait.type="binary")

glmmHeritability(glmm)

seqClose(gdsfile)
}

\keyword{GDS}
\keyword{genetics}
\keyword{association}

\name{seqAssocGLMM_SKAT}
\alias{seqAssocGLMM_SKAT}
\title{SKAT tests}
\description{
    Set-based SKAT tests using mixed models.
}

\usage{
seqAssocGLMM_SKAT(gdsfile, modobj, units, maxMAF=0.01, wbeta=AggrParamBeta,
    missing=0.05, collapse.mac=10, collapse.method=c("max", "sum"),
    ccimb.adj=TRUE, ER.mac=4.5, dsnode="", res.savefn="", res.compress="ZIP",
    parallel=FALSE, verbose=TRUE, verbose.maf=FALSE)
}

\arguments{
    \item{gdsfile}{a SeqArray GDS filename, or a GDS object}
    \item{modobj}{an R object for SAIGE model parameters}
    \item{units}{a list of units of selected variants, with S3 class
        \code{"SeqUnitListClass"} defined in the SeqArray package}
    \item{maxMAF}{a numeric vector for MAF thresholds, e.g.,
        \code{c(0.01, 0.001)} for rare variants}
    \item{wbeta}{weights for per-variant effect, using beta distribution
        \code{dbeta()} according to variant's MAF; a length-two vector, or a
        matrix with two rows for multiple beta parameters; by default, using
        beta(1,1) and beta(1,25) both}
    \item{missing}{missing threshold for variants (checking <= \code{missing}),
        \code{NaN} for no filter}
    \item{collapse.mac}{the mac threshold for collapsing ultra rare variants}
    \item{collapse.method}{maximum of dosages of rare variants (\code{"max"},
        by default), sum up rare genotypes (\code{"sum"})}
    \item{ccimb.adj}{whether adjusting for case-control imbalance}
    \item{ER.mac}{the maximum sum of MAC for the variants with efficient
        resampling p-values}
    \item{dsnode}{"" for automatically searching the GDS nodes "genotype" and
        "annotation/format/DS", or use a user-defined GDS node in the file}
    \item{res.savefn}{an RData or GDS file name, "" for no saving}
    \item{res.compress}{the compression method for the output file, it should
        be one of LZMA, LZMA_RA, ZIP, ZIP_RA and none}
    \item{parallel}{\code{FALSE} (serial processing), \code{TRUE} (multicore
        processing), a numeric value for the number of cores, or other value;
        \code{parallel} is passed to the argument \code{cl} in
        \code{\link[SeqArray]{seqParallel}}, see
        \code{\link[SeqArray]{seqParallel}} for more details}
    \item{verbose}{if \code{TRUE}, show information}
    \item{verbose.maf}{if \code{TRUE}, show summary of MAFs in units}
}

\details{
    \code{seqUnitFilterCond()} in the SeqArray package can be used to restrict
the variant sets \code{units} to a range of MAC and/or MAF.

    For more details of the SAIGE-GENE method, please refer to the SAIGE paper
[Zhou et al. 2020] (see the reference section).
}

\value{
    Return a \code{data.frame} with the following components if not saving to
a file:
    \item{chr}{chromosome;}
    \item{start}{the starting position;}
    \item{end}{the ending position;}
    \item{maxMAF}{maximum of MAFs;}
    \item{numvar}{the number of variants in the window; if
        \code{weight="Cauchy"}, it is the number of individual p-values in
        the Cauchy test}
    \item{macmin}{the minimum of MAC in the window;}
    \item{macmed}{the median of MAC in the window;}
    \item{macmax}{the maximum of MAC in the window;}
    \item{summac}{the sum of minor allele counts;}
    \item{weight}{the weighting strategy with the beta distribution parameters}
    \item{n_collapse}{the number of variants running for single variant tests;}
    \item{g_ncol}{the number of ultra rare variants for collapsing burden
        tests (\code{<= collapse.mac});}
    \item{g_minMAC}{x}
    \item{pval}{SKAT p-value;}
}

\references{
    Zhou W, Zhao Z, Nielsen JB, Fritsche LG, LeFaive J, Taliun SAG, Bi W,
Gabrielsen ME, Daly MJ, Neale BM, Hveem K, Abecasis GR, Willer CJ, et al.
Scalable generalized linear mixed model for region-based association tests
in large biobanks and cohorts. Nat Genet. 2020; 52: 634-9.
}
\author{Xiuwen Zheng}
\seealso{
    \code{\link{seqAssocGLMM_ACAT_V}},
    \code{\link{seqAssocGLMM_ACAT_O}},
    \code{\link{seqAssocGLMM_Burden}},
    \code{\link[SeqArray]{seqUnitFilterCond}}
}

\examples{
# open a GDS file
fn <- system.file("extdata", "grm1k_10k_snp.gds", package="SAIGEgds")
gdsfile <- seqOpen(fn)

sp_grm_fn <- system.file("extdata", "grm1k_10k_sp_grm.rds", package="SAIGEgds")
sp_grm <- readRDS(sp_grm_fn)

# load phenotype
phenofn <- system.file("extdata", "pheno.txt.gz", package="SAIGEgds")
pheno <- read.table(phenofn, header=TRUE, as.is=TRUE)
head(pheno)

# fit the null model
glmm <- seqFitNullGLMM_SPA(y ~ x1 + x2, pheno, gdsfile, sp_grm,
    trait.type="binary")

# get a list of variant units for burden tests
units <- seqUnitSlidingWindows(gdsfile, win.size=50, win.shift=250)

a <- seqAssocGLMM_SKAT(gdsfile, glmm, units, maxMAF=c(0.1,0.01))
head(a)

# close the GDS file
seqClose(gdsfile)
}

\keyword{GDS}
\keyword{genetics}
\keyword{association}
